/*--------------------------------------------------------------------------*/
/*--- ALBERTA:  an Adaptive multi Level finite element toolbox using     ---*/
/*---           Bisectioning refinement and Error control by Residual    ---*/
/*---           Techniques for scientific Applications                   ---*/
/*---                                                                    ---*/
/*--- file: lagrange_2.c                                                 ---*/
/*---                                                                    ---*/
/*--- description: implementation of the basis functions                 ---*/
/*---              lagrange4 in 1d                                       ---*/
/*---                                                                    ---*/
/*--- created by: kgs on host enigma                                     ---*/
/*---           at 16:14 on 28 of March 2003                             ---*/
/*--------------------------------------------------------------------------*/
/*---                                                                    ---*/
/*--- authors:   Alfred Schmidt                                          ---*/
/*---            Zentrum fuer Technomathematik                           ---*/
/*---            Fachbereich 3 Mathematik/Informatik                     ---*/
/*---            Universitaet Bremen                                     ---*/
/*---            Bibliothekstr. 2                                        ---*/
/*---            D-28359 Bremen, Germany                                 ---*/
/*---                                                                    ---*/
/*---            Kunibert G. Siebert                                     ---*/
/*---            Institut fuer Mathematik                                ---*/
/*---            Universitaet Augsburg                                   ---*/
/*---            Universitaetsstr. 14                                    ---*/
/*---            D-86159 Augsburg, Germany                               ---*/
/*---                                                                    ---*/
/*---    http://www.alberta-fem.de/                                      ---*/
/*---                                                                    ---*/
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include <alberta.h>

/*---8<---------------------------------------------------------------------*/
/*---   second order                                                     ---*/
/*--------------------------------------------------------------------->8---*/

/*--------------------------------------------------------------------------*/
/*---  basisfunction 0 located at vertex 0                               ---*/
/*--------------------------------------------------------------------------*/

static REAL phi2_0(const REAL lambda[DIM+1])
{
  return(lambda[0]*(2.0*lambda[0] - 1.0));
}

static const REAL *grd_phi2_0(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0, 0.0};

  grd[0] = 4.0*lambda[0] - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi2_0(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{4.0, 0.0}, {0.0, 0.0}};

  return(D2);
}

/*--------------------------------------------------------------------------*/
/*---  basisfunction 1 located at vertex 1                               ---*/
/*--------------------------------------------------------------------------*/

static REAL phi2_1(const REAL lambda[DIM+1])
{
  return(lambda[1]*(2.0*lambda[1] - 1.0));
}

static const REAL *grd_phi2_1(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0};
  grd[1] = 4.0*lambda[1] - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi2_1(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0.0, 0.0},{0.0, 4.0}};

  return(D2);
}

/*--------------------------------------------------------------------------*/
/*---  basisfunction 2 located at the center                             ---*/
/*--------------------------------------------------------------------------*/

static REAL phi2_2(const REAL lambda[DIM+1])
{
  return(4.0*lambda[0]*lambda[1]);
}

static const REAL *grd_phi2_2(const REAL lambda[DIM+1])
{
  static REAL  grd[DIM+1] = {0.0};

  grd[0] = 4.0*lambda[1];
  grd[1] = 4.0*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi2_2(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0.0, 4.0},{4.0, 0.0}};
  return(D2);
}


/*--------------------------------------------------------------------*/
/*--- function for accessing local DOFs on an element              ---*/
/*--------------------------------------------------------------------*/

static const DOF *get_dof_indices2(const EL *el, const DOF_ADMIN *admin,
                                  DOF *idof)
{
  static DOF  dof_vec[3];
  DOF         *rvec = idof ? idof : dof_vec;
  int         ibas = 0, i, n0, node;
  DOF         **dof = el->dof;

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = admin->mesh->node[VERTEX];
  n0   = admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = dof[node+i][n0];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = admin->mesh->node[CENTER];
  n0   = admin->n0_dof[CENTER];
  rvec[ibas++] = dof[node][n0];


  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing boundary type of DOFs                 ---*/
/*--------------------------------------------------------------------*/

static const S_CHAR *get_bound2(const EL_INFO *el_info, S_CHAR *vec)
{
  FUNCNAME("get_bound2");
  static S_CHAR  my_vec[3];
  S_CHAR         *rvec = vec ? vec : my_vec;
  int            ibas = 0, i;

  TEST_FLAG(FILL_BOUND, el_info);

/*--------------------------------------------------------------------*/
/*--- basis functions at vertices                                  ---*/
/*--------------------------------------------------------------------*/

  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = el_info->bound[i];

/*--------------------------------------------------------------------*/
/*--- basis functions at center                                    ---*/
/*--------------------------------------------------------------------*/

  rvec[ibas++] = INTERIOR;

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- vector of barycentric coordinates                            ---*/
/*--------------------------------------------------------------------*/

static const REAL coords2[3][DIM+1] = 
{
  {1.000000000000000, 0.000000000000000},
  {0.000000000000000, 1.000000000000000},
  {0.500000000000000, 0.500000000000000}
};

/*--------------------------------------------------------------------*/
/*--- function for local interpolaton of scalar functions          ---*/
/*--------------------------------------------------------------------*/

static const REAL *interpol2(const EL_INFO *el_info,
                             int no, const int *b_no,
                             REAL (*f)(const REAL_D),
                             REAL *vec)
{
  FUNCNAME("interpol2");
  static REAL   my_vec[3];
  REAL          *rvec = vec ? vec : my_vec;
  int              i, j;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
     TEST_EXIT(no > 0 && no <= 3)("not for %d points\n", no);
  }

  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    REAL_D     world[3];
    REAL       coord_no[3][DIM+1];
    const REAL (*coord)[DIM+1];

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        for (j = 0; j <= DIM; j++)
          coord_no[i][j] = coords2[b_no[i]][j];
      }
      coord = (const REAL (*)[DIM+1]) coord_no;
    }
    else
    {
       no = 3;
       coord = coords2;
    }

    (*parametric->coord_to_world)(el_info, nil, no, (void *) coord, world);

    for (i = 0; i < no; i++)
      rvec[i] = (*f)(world[i]);
  }
  else
  {
    TEST_FLAG(FILL_COORDS, el_info);

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        const REAL *world = coord_to_world(el_info, coords2[b_no[i]], nil);
        rvec[i] = (*f)(world);
      }
    }
    else
    {
      for (i = 0; i < 3; i++)
      {
        const REAL *world = coord_to_world(el_info, coords2[i], nil);
        rvec[i] = (*f)(world);
      }
    }
  }

  return(rvec);
}
/*--------------------------------------------------------------------*/
/*--- function for local interpolaton of vector functions          ---*/
/*--------------------------------------------------------------------*/

static const REAL_D *interpol_d2(const EL_INFO *el_info,
                                 int no, const int *b_no,
                                 const REAL *(*f)(const REAL_D, REAL_D),
                                 REAL_D *vec)
{
  FUNCNAME("interpol_d2");
  static REAL_D my_vec[3];
  REAL_D        *rvec = vec ? vec : my_vec;
  int              i, j;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
     TEST_EXIT(no > 0 && no <= 3)("not for %d points\n", no);
  }

  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    REAL_D     world[3];
    REAL       coord_no[3][DIM+1];
    const REAL (*coord)[DIM+1];

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        for (j = 0; j <= DIM; j++)
          coord_no[i][j] = coords2[b_no[i]][j];
      }
      coord = (const REAL (*)[DIM+1]) coord_no;
    }
    else
    {
       no = 3;
       coord = coords2;
    }

    (*parametric->coord_to_world)(el_info, nil, no, (void *) coord, world);

    for (i = 0; i < no; i++)
      (*f)(world[i], rvec[i]);
  }
  else
  {
    TEST_FLAG(FILL_COORDS, el_info);

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        const REAL *world = coord_to_world(el_info, coords2[b_no[i]], nil);
        (*f)(world, rvec[i]);
      }
    }
    else
    {
      for (i = 0; i < 3; i++)
      {
        const REAL *world = coord_to_world(el_info, coords2[i], nil);
        (*f)(world, rvec[i]);
      }
    }
  }

  return((const REAL_D *) rvec);
}
/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_INT_VEC                   ---*/
/*--------------------------------------------------------------------*/

static const int *get_int_vec2(const EL *el, const DOF_INT_VEC *dv,
                               int *vec)
{
  FUNCNAME("get_int_vec2");
  static int     my_vec[3];
  int            *rvec = vec ? vec : my_vec;
  int            *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_INT_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  rvec[ibas++] = v[dof[node][n0]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_REAL_VEC                  ---*/
/*--------------------------------------------------------------------*/

static const REAL *get_real_vec2(const EL *el, const DOF_REAL_VEC *dv,
                                 REAL *vec)
{
  FUNCNAME("get_real_vec2");
  static REAL    my_vec[3];
  REAL           *rvec = vec ? vec : my_vec;
  REAL           *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_REAL_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  rvec[ibas++] = v[dof[node][n0]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_REAL_D_VEC                ---*/
/*--------------------------------------------------------------------*/

static const REAL_D *get_real_d_vec2(const EL *el, const DOF_REAL_D_VEC *dv,
                                     REAL_D *vec)
{
  FUNCNAME("get_real_d_vec2");
  static REAL_D  my_vec[3];
  REAL_D         *rvec = vec ? vec : my_vec;
  REAL_D         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_REAL_D_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
  {
    for (n = 0; n < DIM_OF_WORLD; n++)
      rvec[ibas][n] = v[dof[node+i][n0]][n];
    ibas++;
  }

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  for (n = 0; n < DIM_OF_WORLD; n++)
    rvec[ibas][n] = v[dof[node][n0]][n];
  ibas++;

  return((const REAL_D *) rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_SCHAR_VEC                 ---*/
/*--------------------------------------------------------------------*/

static const S_CHAR *get_schar_vec2(const EL *el, const DOF_SCHAR_VEC *dv,
                                    S_CHAR *vec)
{
  FUNCNAME("get_schar_vec2");
  static S_CHAR  my_vec[3];
  S_CHAR         *rvec = vec ? vec : my_vec;
  S_CHAR         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_SCHAR_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  rvec[ibas++] = v[dof[node][n0]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_UCHAR_VEC                 ---*/
/*--------------------------------------------------------------------*/

static const U_CHAR *get_uchar_vec2(const EL *el, const DOF_UCHAR_VEC *dv,
                                    U_CHAR *vec)
{
  FUNCNAME("get_uchar_vec2");
  static U_CHAR  my_vec[3];
  U_CHAR         *rvec = vec ? vec : my_vec;
  U_CHAR         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_UCHAR_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

/*--------------------------------------------------------------------*/
/*--- DOFs at center                                               ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[CENTER];
  n0   = dv->fe_space->admin->n0_dof[CENTER];
  rvec[ibas++] = v[dof[node][n0]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton of DOF_REAL_VECs during refinement ---*/
/*--------------------------------------------------------------------*/

static void refine_inter2(DOF_REAL_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *cdof;
  const REAL      *pvec;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL            *v = dv->vec;

  el = list->el;
  pvec = (*bas_fcts->get_real_vec)(el, dv, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, first child                           ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[0], admin, nil);

  v[cdof[1]] = (pvec[2]);
  v[cdof[2]] = (0.375*pvec[0] - 0.125*pvec[1] + 0.75*pvec[2]);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, second child                          ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[1], admin, nil);

  v[cdof[2]] = (-0.125*pvec[0] + 0.375*pvec[1] + 0.75*pvec[2]);

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton of DOF_REAL_D_VECs during refinement -*/
/*--------------------------------------------------------------------*/

static void refine_inter_d2(DOF_REAL_D_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *cdof;
  const REAL_D    *pvec;
  int             n;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL_D          *v = dv->vec;

  el = list->el;
  pvec = (*bas_fcts->get_real_d_vec)(el, dv, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, first child                           ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[0], admin, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[cdof[1]][n] = (pvec[2][n]);
    v[cdof[2]][n] = (0.375*pvec[0][n] - 0.125*pvec[1][n] + 0.75*pvec[2][n]);
  }

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, second child                          ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[1], admin, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[cdof[2]][n] = (-0.125*pvec[0][n] + 0.375*pvec[1][n] + 0.75*pvec[2][n]);
  }

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton of DOF_REAL_VECs during coarsening ---*/
/*--------------------------------------------------------------------*/

static void coarse_inter2(DOF_REAL_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL      *cvec;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL            *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[0], dv, nil);

  v[pdof[2]] = (cvec[1]);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, second child                          ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[1], dv, nil);

  v[pdof[2]] = (cvec[0]);

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton of DOF_REAL_D_VECs during coarsening -*/
/*--------------------------------------------------------------------*/

static void coarse_inter_d2(DOF_REAL_D_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL_D    *cvec;
  int             n;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL_D          *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[0], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[2]][n] = (cvec[1][n]);
  }

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, second child                          ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[1], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[2]][n] = (cvec[0][n]);
  }

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for restriction of DOF_REAL_VECs during coarsening  ---*/
/*--------------------------------------------------------------------*/

static void coarse_restr2(DOF_REAL_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL      *cvec;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL            *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[0], dv, nil);

  v[pdof[0]] += (0.375*cvec[2]);
  v[pdof[1]] += (-0.125*cvec[2]);
  v[pdof[2]]  = (cvec[1] + 0.75*cvec[2]);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from second child                     ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[1], dv, nil);

  v[pdof[0]] += (-0.125*cvec[2]);
  v[pdof[1]] += (0.375*cvec[2]);
  v[pdof[2]] += (0.75*cvec[2]);

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for restriction of DOF_REAL_D_VECs during coarsening --*/
/*--------------------------------------------------------------------*/

static void coarse_restr_d2(DOF_REAL_D_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL_D    *cvec;
  int             n;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL_D          *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[0], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[0]][n] += (0.375*cvec[2][n]);
    v[pdof[1]][n] += (-0.125*cvec[2][n]);
    v[pdof[2]][n]  = (cvec[1][n] + 0.75*cvec[2][n]);
  }

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from second child                     ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[1], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[0]][n] += (-0.125*cvec[2][n]);
    v[pdof[1]][n] += (0.375*cvec[2][n]);
    v[pdof[2]][n] += (0.75*cvec[2][n]);
  }

  return;
}

/*--------------------------------------------------------------------*/
/*--- Collect all information about basis functions                ---*/
/*--------------------------------------------------------------------*/

static BAS_FCT *phi2[3] =
{
  phi2_0, phi2_1, phi2_2
};

static GRD_BAS_FCT *grd_phi2[3] =
{
  grd_phi2_0, grd_phi2_1, grd_phi2_2
};
static D2_BAS_FCT *D2_phi2[3] =
{
  D2_phi2_0, D2_phi2_1, D2_phi2_2
};

static BAS_FCTS lagrange2 =
{
  "lagrange2", 3, 2, {1, 1},
  nil, phi2, grd_phi2, D2_phi2,
  get_dof_indices2,
  get_bound2,
  interpol2,
  interpol_d2,
  get_int_vec2,
  get_real_vec2,
  get_real_d_vec2,
  get_uchar_vec2,
  get_schar_vec2,
  refine_inter2,
  coarse_inter2,
  coarse_restr2,
  refine_inter_d2,
  coarse_inter_d2,
  coarse_restr_d2,
  (void *) coords2
};
