/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     macro.c                                                        */
/*                                                                          */
/*                                                                          */
/* description:  dimension dependent part of reading/writing macro          */
/*               triangulations for 1d                                      */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*             Daniel Koester                                               */
/*             Institut fuer Mathematik                                     */
/*             Albert-Ludwigs-Universitaet Freiburg                         */
/*             Hermann-Herder-Str. 10                                       */
/*             D-79104 Freiburg                                             */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include <ctype.h>
#include <rpc/types.h>
#include <rpc/xdr.h>
#include "alberta.h"

/*--------------------------------------------------------------------------*/
/* these definitions are necessary at the moment for the macros "MEM_ALLOC" */
/* and "MEM_FREE" to work properly                                          */
/*--------------------------------------------------------------------------*/

typedef int INT_NVERT[N_VERTICES];
typedef int INT_NNEIGH[N_NEIGH];
typedef S_CHAR SCHAR_NNEIGH[N_NEIGH];
typedef U_CHAR UCHAR_NNEIGH[N_NEIGH];

/*--------------------------------------------------------------------------*/
/* read data->bound into mel[].bound[]                                      */
/*--------------------------------------------------------------------------*/

static void fill_bound_info(MESH *mesh, MACRO_DATA *data,
                            const BOUNDARY *(*bdry)(MESH *, int))
{
  int i;
  MACRO_EL   *mel = mesh->first_macro_el;

  for(i = 0; i < data->n_macro_elements; i++) {
    mel[i].bound[1] = data->boundary[i][0];
    mel[i].bound[0] = data->boundary[i][1];
  }

  return;
}


/*--------------------------------------------------------------------------*/
/* orientation(): checks and corrects whether the element is oriented from  */
/* left to right.                                                           */
/*--------------------------------------------------------------------------*/

ALBERTA_DEFUNUSED(static U_CHAR orientation(MACRO_DATA *data))
{
  int      i, vert_buffer, neigh_buffer;
  S_CHAR   bound_buffer;
  U_CHAR   result = false;

  for(i = 0; i < data->n_macro_elements; i++) {
    if(data->coords[data->mel_vertices[i][0]][0] >
       data->coords[data->mel_vertices[i][1]][0]) {
      result = true;

      vert_buffer = data->mel_vertices[i][0];
      data->mel_vertices[i][0] = data->mel_vertices[i][1];
      data->mel_vertices[i][1] = vert_buffer;

      neigh_buffer = data->neigh[i][0];
      data->neigh[i][0] = data->neigh[i][1];
      data->neigh[i][1] = neigh_buffer;

      bound_buffer = data->boundary[i][0];
      data->boundary[i][0] = data->boundary[i][1];
      data->boundary[i][1] = bound_buffer;
    }
  }

  return(result);
}


/*--------------------------------------------------------------------------*/
/* macro_test(): In 1D only element orientation is checked (and corrected). */
/*--------------------------------------------------------------------------*/

void macro_test(MACRO_DATA *data, const char *new_name)
{
  FUNCNAME("macro_test");
 
  U_CHAR error_found = false;

#if DIM_OF_WORLD==1
  if(orientation(data)) {
    error_found = true;
    WARNING("Element orientation was corrected for some elements.\n");
  }
#endif

  if (error_found && new_name) {
    MSG("Attempting to write corrected macro data to file %s...\n", new_name);
    write_macro_data(data, new_name);
  }

  return;
}


#if 0   /* might be needed, if more DOFs are needed per element!!!!   */

/*--------------------------------------------------------------------------*/
/*  boundary_more_boundary_dofs():                                          */
/*  adds dof's at each edge, counts the edges                               */
/*  function pointer bdry only used in 3d!                                  */
/*--------------------------------------------------------------------------*/

static void fill_more_boundary_dofs(MESH *mesh,
                                    const BOUNDARY *(*bdry)(MESH *, int ))
{
  int        i, n_edges = 0, node = mesh->node[EDGE];
  MACRO_EL   *mel;
  DOF        *dof;

  for (mel = mesh->first_macro_el; mel; mel = mel->next)
  {
    for (i = 0; i < N_NEIGH; i++)
    {
      if (!mel->neigh[i] || (mel->neigh[i]->index < mel->index))
      {
	n_edges++;
	if (mesh->n_dof[EDGE])
	{
	  dof = mel->el->dof[node+i] = get_dof(mesh, EDGE);
	  if (mel->neigh[i])
	    mel->neigh[i]->el->dof[node+mel->opp_vertex[i]] = dof;
	}
      }  
    }
  }
  
  mesh->n_edges = n_edges;

  return;
}

#endif

#include "macro_common.c"

























