/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     graphXO.c                                                      */
/*                                                                          */
/*                                                                          */
/* description:  simple graphical routines in 2d                            */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

#if HAVE_LIBGL
#include <GL/glx.h>
#include <GL/gl.h>
#include <unistd.h>
#endif

#include "PLOTansi.h"

/*---8<---------------------------------------------------------------------*/
/*---  standard colors                                                   ---*/
/*--------------------------------------------------------------------->8---*/

const GRAPH_RGBCOLOR rgb_black   = {0.0,0.0,0.0};
const GRAPH_RGBCOLOR rgb_white   = {1.0,1.0,1.0};
const GRAPH_RGBCOLOR rgb_red     = {1.0,0.0,0.0};
const GRAPH_RGBCOLOR rgb_green   = {0.0,1.0,0.0};
const GRAPH_RGBCOLOR rgb_blue    = {0.0,0.0,1.0};
const GRAPH_RGBCOLOR rgb_yellow  = {1.0,1.0,0.0};
const GRAPH_RGBCOLOR rgb_magenta = {1.0,0.0,1.0};
const GRAPH_RGBCOLOR rgb_cyan    = {0.0,1.0,1.0};
const GRAPH_RGBCOLOR rgb_grey50  = {0.5,0.5,0.5};

const GRAPH_RGBCOLOR rgb_albert  = {0.433,0.781,0.969};
const GRAPH_RGBCOLOR rgb_alberta = {0.000,0.391,0.605};


#if HAVE_LIBGL

struct ogl_window
{
  Display      *dpy;
  Window       win;
  GLXContext   context;
  unsigned int width, height;

  Bool         doubleBuffered;
};
typedef struct ogl_window OGL_WINDOW;

static Bool opengl_available = True;

/****************************************************************************/

static Bool WaitForNotify(Display *d, XEvent *e, char *arg)
{
  return (e->type == MapNotify) && (e->xmap.window == (Window)arg);
}

/****************************************************************************/

static Bool OGL_set_std_window(OGL_WINDOW *ogl_win)
{
  if (ogl_win) 
    return(glXMakeCurrent(ogl_win->dpy, ogl_win->win, ogl_win->context));
  else
    return(false);
}

#define OGL_FLUSH(ogl)\
if ((ogl)->doubleBuffered)glXSwapBuffers((ogl)->dpy, ogl->win);else glFlush()

static int open_info = 0;

static XVisualInfo *findVisual(Display *dpy, int screen, Bool *doubleBuffered)
{
	/*
	 * Find a GLX Visual to use.
	 *   Order of preference
	 *       24 bit mono double buffered
	 *       24 bit single buffered
	 *       any depth double buffered
	 *       any depth single buffered
	 */
  int dblBuf24[] = { GLX_RGBA, GLX_DOUBLEBUFFER,
		     GLX_RED_SIZE, 8, GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8, None };
  int snglBuf24[] = { GLX_RGBA, 
		      GLX_RED_SIZE, 8, GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8, None };
  int dblBuf[] = { GLX_RGBA, GLX_DOUBLEBUFFER, None };
  int snglBuf[] = { GLX_RGBA, None };
  
  XVisualInfo *vi = NULL;

  if ((vi = glXChooseVisual(dpy, screen, dblBuf24)))
  {
    INFO(open_info,2)("Using dblBuf24\n");
    *doubleBuffered = True;
    return(vi);
  } 

  if ((vi = glXChooseVisual(dpy, screen, snglBuf24)))
  {
    *doubleBuffered = False;
    INFO(open_info,2)("Using snglBuf24\n");
    return(vi);
  }

  if ((vi = glXChooseVisual(dpy, screen, dblBuf)))
  {
    *doubleBuffered = True;
    INFO(open_info,2)("Using dblBuf\n");
    return(vi);
  }

  if ((vi = glXChooseVisual(dpy, screen, snglBuf)))
  {
    INFO(open_info,2)("Using snglBuf\n");
    *doubleBuffered = False;
    return(vi);
  }

  WARNING("can't find visual\n");
  return(nil);
}

static Window createWindow(Display *dpy, int screen, XVisualInfo *vi,
			   const char *name, const char *geom)
{
  XSetWindowAttributes attributes;
  unsigned long valuemask = 0;
  Window        root = RootWindow(dpy, screen);
  Atom	        delete_window_atom = XInternAtom(dpy, "WM_DELETE_WINDOW", 0);
  Window        window;
  XSizeHints    size_hints;
  XEvent        event;
  int           status;
  const char    *default_geom = "300x300+0+0";

  attributes.colormap = XCreateColormap(dpy, root, vi->visual, AllocNone);
  if (!attributes.colormap)
  {
    WARNING("no color map\n");
    return(0);
  }
  valuemask = CWColormap;
  valuemask = CWColormap|CWBorderPixel|CWBackPixel;

  size_hints.flags = PResizeInc | PMinSize ;
  size_hints.width_inc  =  1;
  size_hints.height_inc =  1;
  size_hints.min_width  = 10;
  size_hints.min_height = 10;

  if (!geom) geom = default_geom;

  status = XGeometry(dpy, vi->screen, (char *) geom, (char *) default_geom, 
		     2, 1, 1, 0, 0, &(size_hints.x), &(size_hints.y),
		     &(size_hints.width), &(size_hints.height));
  if (status & (XValue|YValue)) 
  {
    size_hints.flags |= USPosition;
    size_hints.flags &= ~PPosition;
  }
  if (status & (WidthValue|HeightValue)) 
  {
    size_hints.flags |= USSize;
    size_hints.flags &= ~PSize;
  }

  if (size_hints.flags & USSize) 
  {
    size_hints.flags |= PAspect;
    size_hints.min_aspect.x = size_hints.width;
    size_hints.min_aspect.y = size_hints.height;
    size_hints.max_aspect.x = size_hints.width;
    size_hints.max_aspect.y = size_hints.height;
  }

  attributes.border_pixel = 0;
  attributes.background_pixel = 0;
  attributes.event_mask = StructureNotifyMask;
  window = XCreateWindow(dpy, root, size_hints.x, size_hints.y,
			 size_hints.width, size_hints.height, 0, vi->depth,
			 InputOutput, vi->visual, valuemask, &attributes);

  if (!window) 
  {
    WARNING("couldn't create a window\n");
    return(0);
  }

  XSetStandardProperties(dpy, window, name, name, None, NULL, 0, &size_hints);
  XSelectInput(dpy, window, ExposureMask|StructureNotifyMask);

  XMapWindow(dpy, window);
  
  XSetWMProtocols(dpy, window, &delete_window_atom, 1);
  XIfEvent(dpy, &event, WaitForNotify, (char*) window);
  XStoreName(dpy, window, name);

  return(window);
}

static OGL_WINDOW *OGL_create_window(const char *name, const char *geometry)
{
  FUNCNAME("OGL_create_window");
  Display      *dpy;
  int          screen;
  Window       window, root;
  int          dummy;
  Bool	       doubleBuffered;
  XVisualInfo  *vi;
  GLXContext   context;
  unsigned int bwidth, depth;
  int          x, y;
  OGL_WINDOW * ogl_win;
  
  if (!(dpy = XOpenDisplay(NULL))) 
  {
    WARNING("can't open X display\n");
    return (nil);
  }
  screen = DefaultScreen(dpy);

  if (!(opengl_available = glXQueryExtension(dpy, &dummy, &dummy)))
  {
    WARNING("server doesn't support GLX Extension\n");
    return (nil);
  }

  if (!(vi = findVisual(dpy, screen, &doubleBuffered)))
    return(nil);
	
  if (!(window = createWindow(dpy, screen, vi, name, geometry)))
    return(nil);

  context = glXCreateContext(dpy, vi, NULL, GL_TRUE);

  XFree(vi);

  if (context == NULL) 
  {
    WARNING("can't create context\n");
    return(nil);
  }
  
  if (!glXMakeCurrent(dpy, window, context)) 
  {
    WARNING("glXMakeCurrent failed\n");
    return (nil);
  }

  ogl_win = MEM_ALLOC(1, OGL_WINDOW);
  ogl_win->dpy = dpy;
  ogl_win->win = window;
  ogl_win->context  = context;
  ogl_win->doubleBuffered = doubleBuffered;
  
  XGetGeometry(dpy, window, &root, &x, &y,
	       &ogl_win->width, &ogl_win->height,
	       &bwidth, &depth);
  
  return(ogl_win);
}

/****************************************************************************/

static void OGL_destroy_window(OGL_WINDOW *ogl_win)
{
  /* FUNCNAME("OGL_destroy_window"); re-enable when MSG() is used here ... */ 

  if (!ogl_win)  return;

  glXDestroyContext(ogl_win->dpy, ogl_win->context);

  XDestroyWindow(ogl_win->dpy, ogl_win->win);
  XFlush(ogl_win->dpy);

  MEM_FREE(ogl_win, 1, OGL_WINDOW);
}
#endif /*HAVE_LIBGL*/

/****************************************************************************/
static const REAL vertex_lambda[N_VERTICES][DIM+1] = {{1.0,0.0,0.0},
						      {0.0,1.0,0.0},
						      {0.0,0.0,1.0}};

#if 0
static const REAL midedge_lambda[N_VERTICES][DIM+1] = {{0.0,0.5,0.5},
						       {0.5,0.0,0.5},
						       {0.5,0.5,0.0}};
#endif

static float xmin[DIM_OF_WORLD], xmax[DIM_OF_WORLD], diam[DIM_OF_WORLD];
static void xminmax_fct(const EL_INFO *elinfo)
{
  FUNCNAME("xminmax_fct");
  int i, j;
  PARAMETRIC *parametric = elinfo->mesh->parametric;

  if (parametric && parametric->init_element(elinfo, parametric)) {
    REAL_D world[N_VERTICES];

    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, world);

    for (i = 0; i < N_VERTICES; i++) {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xmin[j] = MIN(xmin[j], world[i][j]);
	xmax[j] = MAX(xmax[j], world[i][j]);
      }
    }
  }
  else {
    TEST_FLAG(FILL_COORDS, elinfo);
    for (i = 0; i < N_VERTICES; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xmin[j] = MIN(xmin[j], elinfo->coord[i][j]);
	xmax[j] = MAX(xmax[j], elinfo->coord[i][j]);
      }    
  }
  return;
}


GRAPH_WINDOW graph_open_window(const char *title, const char *geometry,
			       REAL *world, MESH *mesh)
{
#if HAVE_LIBGL
  OGL_WINDOW *winO = nil;
#endif
  PLOT_INFO  *winX = nil;
  char       Geometry[16];
  int        i;
 
  if (world)
  {
    xmin[0] = world[0];
    xmax[0] = world[1];
    xmin[1] = world[2];
    xmax[1] = world[3];
    diam[0] = MAX(xmax[0] - xmin[0], 1.E-10);
    diam[1] = MAX(xmax[1] - xmin[1], 1.E-10);
  }
  else if (mesh)
  {
    for (i=0; i<DIM_OF_WORLD; i++)
      xmax[i] = -(xmin[i] = 1.0E10);
    mesh_traverse(mesh, -1, CALL_LEAF_EL|FILL_COORDS, xminmax_fct);
    for (i=0; i<DIM_OF_WORLD; i++)
    {
      diam[i] = MAX(xmax[i] - xmin[i], 1.E-10);
      xmin[i] -= 0.1 * diam[i];
      xmax[i] += 0.1 * diam[i];
      diam[i] *= 1.2;
    }
  }
  else
  {
    for (i=0; i<DIM_OF_WORLD; i++)
    {
      xmin[i] = 0.0;
      xmax[i] = 1.0;
      diam[i] = 1.0;
    }
  }

  if (!title)
    title = "ALBERTA graphics";

  if (!geometry)
  {
    static int SIZE = 400, xsize, ysize;

    if (diam[0] >= diam[1])
    {
      xsize = SIZE;
      ysize = SIZE * diam[1] / diam[0];
    }
    else
    {
      xsize = SIZE * diam[0] / diam[1];
      ysize = SIZE;
    }
    sprintf(Geometry, "%dx%d", xsize, ysize);
    geometry = Geometry;
/*     MSG("use geometry: %s\n", geometry); */
  }

#if HAVE_LIBGL
  if (opengl_available)
    winO = OGL_create_window(title, geometry);
  if (!opengl_available)
    winX = PLOT_create_window(title, geometry);
#else
  winX = PLOT_create_window(title, geometry);
#endif

#if HAVE_LIBGL
  if (winO) {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glOrtho(xmin[0], xmax[0], xmin[1], xmax[1], -1.0, 1.0);

    graph_clear_window((GRAPH_WINDOW) winO, rgb_white);
    return((GRAPH_WINDOW) winO);
  }
  else
#endif
  {
    /*PLOT_world2d(xmin[0], xmax[0], xmin[1], xmax[1]);*/
    if (winX->windowWidth > winX->windowHeight)
    {
      PLOT_world2d(xmin[0], 
		   xmin[0] + (diam[0]*winX->windowHeight)/winX->windowWidth,
		   xmin[1], xmax[1]);
    }
    else
    {
      PLOT_world2d(xmin[0], xmax[0], xmin[1],
		   xmin[1] + (diam[1]*winX->windowWidth)/winX->windowHeight);
    }
    PLOT_set_backgroundcolor((float *) rgb_white);
    PLOT_clear();
    PLOT_flush();

    return((GRAPH_WINDOW) winX);
  }
}

/****************************************************************************/

static FLAGS graph_mesh_flags;

static const float *linecolor;
/****************************************************************************/
#if HAVE_LIBGL

/****************************************************************************/
static void OGL_clear_window(OGL_WINDOW *ogl_win, const GRAPH_RGBCOLOR c)
{
  FUNCNAME("OGL_clear_window");

  Window       root;
  unsigned int w, h, depth, bwidth;
  int          x, y;

  if (!ogl_win) {
    MSG("no OGL_WINDOW\n");
    return;
  }

  OGL_set_std_window(ogl_win);

  /* Has the window size changed? */
  XGetGeometry(ogl_win->dpy, ogl_win->win,
	       &root, &x, &y, &w, &h, &bwidth, &depth);
 
  if (w != ogl_win->width || h != ogl_win->height){
    ogl_win->width  = w;
    ogl_win->height = h;
    XResizeWindow(ogl_win->dpy, ogl_win->win, w, h);
  }

  glXMakeCurrent(ogl_win->dpy, ogl_win->win, ogl_win->context);

  glViewport(0, 0, w, h);
  if (c)
    glClearColor(c[0], c[1], c[2], 1.0);
  else
    glClearColor(1.0, 1.0, 1.0, 1.0);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  OGL_FLUSH(ogl_win);
}

/****************************************************************************/


static void OGLgraph_mesh_fct(const EL_INFO *elinfo)
{
  FUNCNAME("OGLgraph_mesh_fct");

  EL       *el = elinfo->el;
  int      i, j, i1,i2;
  double   xmid[DIM_OF_WORLD];
  DOF      **dof;
  char     cindex[20];
  REAL_D   world[N_VERTICES];
  const REAL_D   *coord;
  PARAMETRIC *parametric = elinfo->mesh->parametric;

  if (parametric && parametric->init_element(elinfo, parametric)) {
    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, world);
    coord = (const REAL_D *) world;
  }
  else {
    TEST_FLAG(FILL_COORDS, elinfo);
    coord = elinfo->coord;
  }

  for (j=0; j<DIM_OF_WORLD;j++) {
    xmid[j] = 0.0;
    for (i=0;i<N_VERTICES;i++) xmid[j] += coord[i][j];
    xmid[j] /= N_VERTICES;
  }

  /* show refinement/coarsening mark */
  if (graph_mesh_flags & GRAPH_MESH_ELEMENT_MARK)
  {
    if      (el->mark > 0) glColor3fv(rgb_red);
    else if (el->mark < 0) glColor3fv(rgb_blue);
    else                   glColor3fv(rgb_white);

    glBegin(GL_TRIANGLES);
    glVertex2dv(coord[0]);
    glVertex2dv(coord[1]);
    glVertex2dv(coord[2]);
    glEnd();
  }


#if EL_INDEX
  /* show element indices */
  if (graph_mesh_flags & GRAPH_MESH_ELEMENT_INDEX)
  {
    if (elinfo->level < 5) {
      sprintf(cindex,"%d",el->index);
      /* PLOT_text2d(xmid[0],xmid[1],cindex); */
    }
  }
#endif


  if (graph_mesh_flags & GRAPH_MESH_BOUNDARY)
  {
    /* show boundary edges */
    TEST_FLAG(FILL_BOUND, elinfo);

    for (i=0; i<N_NEIGH; i++) {
      if (!IS_INTERIOR(elinfo->boundary[i]))
      {
	if (linecolor)
	  glColor3fv(linecolor);
	else if (IS_DIRICHLET(elinfo->boundary[i]))
	  glColor3fv(rgb_blue);
	else
	  glColor3fv(rgb_red);

	i1 = (i+1) % 3;
	i2 = (i+2) % 3;
	
	glBegin(GL_LINE_STRIP);
	glVertex2dv(coord[i1]);
	/* if (parametric) insert additional points... */
	glVertex2dv(coord[i2]);
	glEnd();
      }
    }
  }
  else
  {
    /* show all edges */
    glColor3fv(linecolor ? linecolor : rgb_black);
    glBegin(GL_LINE_LOOP);
    for (i=0;i<N_VERTICES;i++) {
      glVertex2dv(coord[i]);
      /* if (parametric) insert additional points... */
    }
    glEnd();
  }

  /* show dof[0] at vertices */
  if (graph_mesh_flags & GRAPH_MESH_VERTEX_DOF)
  {
    if ((dof = el->dof)) 
    {
      for (i=0; i<N_VERTICES; i++) 
      {
	if (dof[i]) sprintf(cindex,"%d",dof[i][0]);
      }
    }
  }
  return;
}
#endif

static void Xgraph_mesh_fct(const EL_INFO *elinfo)
{
  FUNCNAME("Xgraph_mesh_fct");

  EL       *el = elinfo->el;
  int      i, j, i1, i2;
  float    xmid[DIM_OF_WORLD], tx[N_VERTICES][DIM_OF_WORLD];
  DOF      **dof;
  char     cindex[20];
  REAL_D   world[N_VERTICES];
  const REAL_D   *coord;
  PARAMETRIC *parametric = elinfo->mesh->parametric;

  if (parametric && parametric->init_element(elinfo, parametric)) {
    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, world);
    coord = (const REAL_D *) world;
  }
  else {
    TEST_FLAG(FILL_COORDS, elinfo);
    coord = elinfo->coord;
  }

  for (j=0; j<DIM_OF_WORLD;j++) {
    xmid[j] = 0.0;
    for (i=0;i<N_VERTICES;i++) xmid[j] += coord[i][j];
    xmid[j] /= N_VERTICES;
  }


  for (i=0;i<N_VERTICES;i++) {
    for (j=0; j<DIM_OF_WORLD; j++) tx[i][j] = coord[i][j];
  }

  /* show refinement/coarsening mark */
  if (graph_mesh_flags & GRAPH_MESH_ELEMENT_MARK)
  {
    if      (el->mark > 0) PLOT_set_patchcolor((float *) rgb_red);
    else if (el->mark < 0) PLOT_set_patchcolor((float *) rgb_blue);
    else                   PLOT_set_patchcolor((float *) rgb_white);
    PLOT_vtriangle2d(tx[0], tx[1], tx[2]);
  }


#if EL_INDEX
  /* show element indices */
  if (graph_mesh_flags & GRAPH_MESH_ELEMENT_INDEX)
  {
    if (elinfo->level < 5) {
      sprintf(cindex,"%d",el->index);
      PLOT_text2d(xmid[0],xmid[1],cindex);
    }
  }
#endif


  if (graph_mesh_flags & GRAPH_MESH_BOUNDARY)
  {
    /* show boundary edges */
    TEST_FLAG(FILL_BOUND, elinfo);

    for (i = 0; i < N_NEIGH; i++) 
    {
      if (!IS_INTERIOR(elinfo->boundary[i]))
      {
	if (linecolor)
	  PLOT_set_linecolor(linecolor);
	else if (IS_DIRICHLET(elinfo->boundary[i]))
	  PLOT_set_linecolor((float *) rgb_blue);
	else
	  PLOT_set_linecolor((float *) rgb_red);

	i1 = (i+1) % 3;
	i2 = (i+2) % 3;
	PLOT_move2d(tx[i1][0], tx[i1][1]);
	/* if (parametric) insert additional points... */
	PLOT_draw2d(tx[i2][0], tx[i2][1]);
      }
    }
  }
  else
  {
    /* show all edges */
    PLOT_set_linecolor((float *) (linecolor ? linecolor : rgb_black));

    PLOT_move2d(tx[N_VERTICES-1][0], tx[N_VERTICES-1][1]);
    for (i=0;i<N_VERTICES;i++) {
      PLOT_draw2d(tx[i][0], tx[i][1]);
      /* if (parametric) insert additional points... */
    }
  }

  /* show dof[0] at vertices */
  if (graph_mesh_flags & GRAPH_MESH_VERTEX_DOF)
  {
    if ((dof = el->dof))
    {
      for (i=0; i<N_VERTICES; i++)
      {
	if (dof[i])
	{
	  sprintf(cindex,"%d",dof[i][0]);
	  PLOT_text2d(tx[i][0], tx[i][1], cindex);
	}
      }
    }
  }
  PLOT_flush();
}


void  graph_mesh(GRAPH_WINDOW win, MESH *mesh,
		 const GRAPH_RGBCOLOR c, FLAGS flags)
{
  if (!mesh) return;

/*   if (mesh->parametric)    MSG("parametric mesh!\n"); */

  linecolor = c;
  graph_mesh_flags = flags;

#if HAVE_LIBGL
  if (opengl_available) 
  {
    OGL_WINDOW *ogl_win = (OGL_WINDOW *) win;
    OGL_set_std_window(ogl_win);

    glLineWidth(1.0);

    mesh_traverse(mesh, -1, CALL_LEAF_EL | FILL_COORDS | FILL_BOUND,
		  OGLgraph_mesh_fct);
    OGL_FLUSH(ogl_win);
    return;
  }
#endif

  PLOT_set_std_window((PLOT_INFO *)win);

  mesh_traverse(mesh, -1, CALL_LEAF_EL | FILL_COORDS | FILL_BOUND,
		Xgraph_mesh_fct);

  PLOT_flush();

}

/****************************************************************************/

void  graph_close_window(GRAPH_WINDOW win)
{
#if HAVE_LIBGL
  if (opengl_available) {
    OGL_destroy_window((OGL_WINDOW *)win);
    return;
  }
#endif
  PLOT_destroy_window((PLOT_INFO *)win);
}

void  graph_clear_window(GRAPH_WINDOW win, const GRAPH_RGBCOLOR c)
{
#if HAVE_LIBGL
  if (opengl_available)
  {
    OGL_clear_window((OGL_WINDOW *)win, c ? c : rgb_white);
    return;
  }
#endif

  PLOT_set_std_window((PLOT_INFO *)win);
  PLOT_set_backgroundcolor((float *) (c ? c : rgb_white));
  PLOT_clear();
  PLOT_flush();
}

void graph_point(GRAPH_WINDOW win, const REAL p0[2], const GRAPH_RGBCOLOR c,
		 float ps)
{
#if HAVE_LIBGL
  if (opengl_available)
  {
    OGL_WINDOW *ogl_win = win;
    OGL_set_std_window(ogl_win);

    glColor3fv(c ? c : rgb_black);
    glPointSize(ps > 0 ? ps : 1.0);
    
    glBegin(GL_POINTS);
    glVertex2d(p0[0], p0[1]);
    glEnd();

    OGL_FLUSH(ogl_win);
  }
#endif
  return;
}

/****************************************************************************/
void graph_line(GRAPH_WINDOW win, const REAL p0[2], const REAL p1[2], 
		const GRAPH_RGBCOLOR c, float lw)
{
  float xy[2];
  int   j;

#if HAVE_LIBGL
  if (opengl_available)
  {
    OGL_WINDOW *ogl_win = win;

    OGL_set_std_window(ogl_win);

    glColor3fv(c ? c : rgb_black);
    glLineWidth(lw > 0 ? lw : 1.0);
    
    glBegin(GL_LINE_STRIP);
    for (j=0; j<2; j++) xy[j] = p0[j];
    glVertex2fv(xy);
    for (j=0; j<2; j++) xy[j] = p1[j];
    glVertex2fv(xy);
    glEnd();
    
    OGL_FLUSH(ogl_win);
  }
  else
#endif
  {
    PLOT_set_std_window((PLOT_INFO *)win);
    PLOT_set_linecolor((float *) (c ? c : rgb_black));
    
    for (j=0; j<2; j++) xy[j] = p0[j];
    PLOT_vmove2d(xy);
    for (j=0; j<2; j++) xy[j] = p1[j];
    PLOT_vdraw2d(xy);
    
    PLOT_flush();
  }
  return;
}

/****************************************************************************/
/****************************************************************************/



static REAL level_value = 0.0;
static int  nrefine = 0;

static const DOF_REAL_VEC   *drv  = nil;
static const DOF_REAL_D_VEC *drdv  = nil;

static const BAS_FCTS *bas_fcts = nil;
static int n_bas_fcts     = 0;
static BAS_FCT **phi      = nil;
static const REAL    *el_vec = nil;
static const REAL_D  *el_vec_d = nil;
static PARAMETRIC    *el_parametric = nil;
static const EL_INFO *el_info = nil;

static void graph_level_recursive(int refine, const REAL *b[3],
				  REAL v[3], const REAL *x[3])
{
  int    i,j;

  if (refine > 0)       /* refine and call recursively */
  {
    const REAL *bnew[3], *xnew[3];
    REAL       bm[3][DIM+1], vnew[3], vm[3], xm[3][DIM_OF_WORLD];

    for (j = 0; j <= DIM; j++)
    {
      bm[0][j] = 0.5*(b[1][j] + b[2][j]);
      bm[1][j] = 0.5*(b[0][j] + b[2][j]);
      bm[2][j] = 0.5*(b[0][j] + b[1][j]);
    }

    for (i = 0; i < 3; i++)
    {
      vm[i] = 0.0;
      for (j=0; j<n_bas_fcts; j++)
	vm[i] += el_vec[j] * phi[j](bm[i]);
    }

    if (el_parametric)
    {
      el_parametric->coord_to_world(el_info, nil, N_EDGES,
				    (const REAL (*)[DIM+1])bm, xm);
    }
    else
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xm[0][j] = 0.5*(x[1][j] + x[2][j]);
	xm[1][j] = 0.5*(x[0][j] + x[2][j]);
	xm[2][j] = 0.5*(x[0][j] + x[1][j]);
      }
    }

    bnew[0] = b[0]; bnew[1] = bm[2]; bnew[2] = bm[1]; 
    xnew[0] = x[0]; xnew[1] = xm[2]; xnew[2] = xm[1]; 
    vnew[0] = v[0]; vnew[1] = vm[2]; vnew[2] = vm[1]; 
    graph_level_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[2]; bnew[2] = bm[1]; 
    xnew[0] = xm[2]; xnew[1] = x[2]; xnew[2] = xm[1]; 
    vnew[0] = vm[2]; vnew[1] = v[2]; vnew[2] = vm[1]; 
    graph_level_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = b[2]; bnew[1] = bm[2]; bnew[2] = bm[0]; 
    xnew[0] = x[2]; xnew[1] = xm[2]; xnew[2] = xm[0]; 
    vnew[0] = v[2]; vnew[1] = vm[2]; vnew[2] = vm[0]; 
    graph_level_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[1]; bnew[2] = bm[0]; 
    xnew[0] = xm[2]; xnew[1] = x[1]; xnew[2] = xm[0]; 
    vnew[0] = vm[2]; vnew[1] = v[1]; vnew[2] = vm[0]; 
    graph_level_recursive(refine-1, bnew, vnew, xnew);

    return;
  }

  /* draw level line */

  {
    static REAL small = 1.0E-8;
    int   i1, i2, flag[3];
    REAL  s[3];
    float xy[2];

    for (i=0; i<3; i++) {
      i1 = (i+1) % 3;
      if (ABS(v[i1]-v[i]) >= small) {
	s[i] = (level_value - v[i]) / (v[i1]-v[i]);
	flag[i] = (s[i] <= 1.0 && s[i] >= 0.0);
      }
      else {
	flag[i] = 0;
	if (ABS(level_value - v[i]) <= small) {
#if HAVE_LIBGL
	  if (opengl_available)
	  {
	    glBegin(GL_LINE_STRIP);
	    xy[0] = x[i][0];
	    xy[1] = x[i][1];
	    glVertex2fv(xy);
	    xy[0] = x[i1][0];
	    xy[1] = x[i1][1];
	    glVertex2fv(xy);
	    glEnd();
	  }
	  else
#endif
	  {
	    xy[0] = x[i][0];
	    xy[1] = x[i][1];
	    PLOT_vmove2d(xy);
	    xy[0] = x[i1][0];
	    xy[1] = x[i1][1];
	    PLOT_vdraw2d(xy);
	  }
	}
      }
    }

    for (i=0; i<3; i++) {
      i1 = (i+1) % 3;
      if (flag[i] && flag[i1]) {
	i2 = (i+2) % 3;
#if HAVE_LIBGL
	if (opengl_available)
	{
	  glBegin(GL_LINE_STRIP);
	  for (j=0; j<2; j++)
	    xy[j] = x[i][j] + s[i] * (x[i1][j] - x[i][j]);
	  glVertex2fv(xy);
	  for (j=0; j<2; j++)
	    xy[j] = x[i1][j] + s[i1] * (x[i2][j] - x[i1][j]);
	  glVertex2fv(xy);
	  glEnd();
	}
	else
#endif
	{
	  for (j=0; j<2; j++)
	    xy[j] = x[i][j] + s[i] * (x[i1][j] - x[i][j]);
	  PLOT_vmove2d(xy);
	  for (j=0; j<2; j++)
	    xy[j] = x[i1][j] + s[i1] * (x[i2][j] - x[i1][j]);
	  PLOT_vdraw2d(xy);
	}
      }
    }
  }
}

static void graph_level_fct(const EL_INFO *elinfo)
{
  FUNCNAME("graph_level_fct");
  static REAL bnew[3][DIM+1]={{1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,1.0}};
  static const REAL *b[3]={bnew[0],bnew[1],bnew[2]};
  const REAL  *x[3];
  REAL_D      coord[N_VERTICES];
  REAL        v[3];
  EL          *el = elinfo->el;
  int         i, j;
  PARAMETRIC  *parametric = elinfo->mesh->parametric;

  if (bas_fcts->get_real_vec)
    el_vec = bas_fcts->get_real_vec(el, drv, nil);
  else
    ERROR("no bas_fcts->get_real_vec()\n");

  el_info       = elinfo;
  if (parametric && parametric->init_element(elinfo, parametric)) {
    el_parametric = parametric;
    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, coord);
    for (i = 0; i < 3; i++)
      x[i] = coord[i];
  }
  else {
    el_parametric = nil;
    for (i = 0; i < 3; i++)
      x[i] = elinfo->coord[i];
  }

  for (i = 0; i < 3; i++)
  {
    v[i] = 0.0;
    for (j=0; j<n_bas_fcts; j++)
      v[i] += el_vec[j] * phi[j](b[i]);
  }

  graph_level_recursive(nrefine, b, v, x);
}

/****************************************************************************/

void  graph_level(GRAPH_WINDOW win, const DOF_REAL_VEC *v, REAL level,
		  const GRAPH_RGBCOLOR c, int refine)
{
  FUNCNAME("graph_level");
#if HAVE_LIBGL
  OGL_WINDOW *ogl_win = win;
#endif

  if (!v) return;

  if (v->fe_space && v->fe_space->admin && v->fe_space->admin->mesh)
  {
#if HAVE_LIBGL
    if (opengl_available)
    {
      OGL_set_std_window(ogl_win);

      glLineWidth(1);
      glColor3fv(c ? c : rgb_black);
    }
    else
#endif
    {
      PLOT_set_std_window((PLOT_INFO *)win);
      PLOT_set_linecolor((float *) (c ? c : rgb_black));
    }

    bas_fcts     = v->fe_space->bas_fcts;
    n_bas_fcts   = bas_fcts->n_bas_fcts;
    phi          = bas_fcts->phi;
    drv          = v;
    level_value  = level;
    if (refine >= 0)
      nrefine = refine;
    else
      nrefine = MAX(0, bas_fcts->degree-1);

    mesh_traverse(v->fe_space->admin->mesh, -1, CALL_LEAF_EL | FILL_COORDS,
		  graph_level_fct);
    
#if HAVE_LIBGL
    if (opengl_available)
      OGL_FLUSH(ogl_win);
    else
#endif
    {
      PLOT_flush();
    }
  }
  else
    ERROR("no FE_SPACE OR DOF_ADMIN or MESH\n");
}

/****************************************************************************/

static void graph_level_d_recursive(int refine, const REAL *b[3], REAL v[3],
                                    const REAL *x[3])
{
  int    i,j,k;

  if (refine > 0)       /* refine and call recursively */
  {
    const REAL *bnew[3], *xnew[3];
    REAL       bm[3][DIM+1], vnew[3], vm[3], xm[3][DIM_OF_WORLD], phi_b;
    REAL_D     vd;

    for (j = 0; j <= DIM; j++)
    {
      bm[0][j] = 0.5*(b[1][j] + b[2][j]);
      bm[1][j] = 0.5*(b[0][j] + b[2][j]);
      bm[2][j] = 0.5*(b[0][j] + b[1][j]);
    }

    for (i = 0; i < 3; i++)
    {
	SET_DOW(0.0, vd);
	for (j=0; j<n_bas_fcts; j++) {
	    phi_b = phi[j](bm[i]);
	    for (k=0; k<DIM_OF_WORLD; ++k)
		vd[k] += el_vec_d[j][k] * phi_b;
	}
	vm[i] = NORM_DOW(vd);
    }

    if (el_parametric)
    {
      el_parametric->coord_to_world(el_info, nil, N_EDGES,
				    (const REAL (*)[DIM+1])bm, xm);
    }
    else
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xm[0][j] = 0.5*(x[1][j] + x[2][j]);
	xm[1][j] = 0.5*(x[0][j] + x[2][j]);
	xm[2][j] = 0.5*(x[0][j] + x[1][j]);
      }
    }

    bnew[0] = b[0]; bnew[1] = bm[2]; bnew[2] = bm[1]; 
    xnew[0] = x[0]; xnew[1] = xm[2]; xnew[2] = xm[1]; 
    vnew[0] = v[0]; vnew[1] = vm[2]; vnew[2] = vm[1]; 
    graph_level_d_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[2]; bnew[2] = bm[1]; 
    xnew[0] = xm[2]; xnew[1] = x[2]; xnew[2] = xm[1]; 
    vnew[0] = vm[2]; vnew[1] = v[2]; vnew[2] = vm[1]; 
    graph_level_d_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = b[2]; bnew[1] = bm[2]; bnew[2] = bm[0]; 
    xnew[0] = x[2]; xnew[1] = xm[2]; xnew[2] = xm[0]; 
    vnew[0] = v[2]; vnew[1] = vm[2]; vnew[2] = vm[0]; 
    graph_level_d_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[1]; bnew[2] = bm[0]; 
    xnew[0] = xm[2]; xnew[1] = x[1]; xnew[2] = xm[0]; 
    vnew[0] = vm[2]; vnew[1] = v[1]; vnew[2] = vm[0]; 
    graph_level_d_recursive(refine-1, bnew, vnew, xnew);

    return;
  }

  /* draw level line */

  {
    static REAL small = 1.0E-8;
    int   i1, i2, flag[3];
    REAL  s[3];
    float xy[2];

    for (i=0; i<3; i++) {
      i1 = (i+1) % 3;
      if (ABS(v[i1]-v[i]) >= small) {
	s[i] = (level_value - v[i]) / (v[i1]-v[i]);
	flag[i] = (s[i] <= 1.0 && s[i] >= 0.0);
      }
      else {
	flag[i] = 0;
	if (ABS(level_value - v[i]) <= small) {
#if HAVE_LIBGL
	  if (opengl_available)
	  {
	    glBegin(GL_LINE_STRIP);
	    xy[0] = x[i][0];
	    xy[1] = x[i][1];
	    glVertex2fv(xy);
	    xy[0] = x[i1][0];
	    xy[1] = x[i1][1];
	    glVertex2fv(xy);
	    glEnd();
	  }
	  else
#endif
	  {
	    xy[0] = x[i][0];
	    xy[1] = x[i][1];
	    PLOT_vmove2d(xy);
	    xy[0] = x[i1][0];
	    xy[1] = x[i1][1];
	    PLOT_vdraw2d(xy);
	  }
	}
      }
    }

    for (i=0; i<3; i++) {
      i1 = (i+1) % 3;
      if (flag[i] && flag[i1]) {
	i2 = (i+2) % 3;
#if HAVE_LIBGL
	if (opengl_available)
	{
	  glBegin(GL_LINE_STRIP);
	  for (j=0; j<2; j++)
	    xy[j] = x[i][j] + s[i] * (x[i1][j] - x[i][j]);
	  glVertex2fv(xy);
	  for (j=0; j<2; j++)
	    xy[j] = x[i1][j] + s[i1] * (x[i2][j] - x[i1][j]);
	  glVertex2fv(xy);
	  glEnd();
	}
	else
#endif
	{
	  for (j=0; j<2; j++)
	    xy[j] = x[i][j] + s[i] * (x[i1][j] - x[i][j]);
	  PLOT_vmove2d(xy);
	  for (j=0; j<2; j++)
	    xy[j] = x[i1][j] + s[i1] * (x[i2][j] - x[i1][j]);
	  PLOT_vdraw2d(xy);
	}
      }
    }
  }
}

static void graph_level_d_fct(const EL_INFO *elinfo)
{
  FUNCNAME("graph_level_d_fct");
  static REAL bnew[3][DIM+1]={{1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,1.0}};
  static const REAL *b[3]={bnew[0],bnew[1],bnew[2]};
  const REAL  *x[3];
  REAL_D      coord[N_VERTICES], vd;
  REAL        v[3], phi_b;
  EL          *el = elinfo->el;
  int         i, j, k;
  PARAMETRIC  *parametric = elinfo->mesh->parametric;

  if (bas_fcts->get_real_d_vec)
    el_vec_d = bas_fcts->get_real_d_vec(el, drdv, nil);
  else
    ERROR("no bas_fcts->get_real_d_vec()\n");

  el_info       = elinfo;
  if (parametric && parametric->init_element(elinfo, parametric)) {
    el_parametric = parametric;
    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, coord);
    for (i = 0; i < 3; i++)
      x[i] = coord[i];
  }
  else {
    el_parametric = nil;
    for (i = 0; i < 3; i++)
      x[i] = elinfo->coord[i];
  }

  for (i = 0; i < 3; i++)
  {
    SET_DOW(0.0, vd);
    for (j=0; j<n_bas_fcts; j++) {
	phi_b = phi[j](b[i]);
	for (k=0; k<DIM_OF_WORLD; ++k)
	    vd[k] += el_vec_d[j][k] * phi_b;
    }
    v[i] = NORM_DOW(vd);
  }

  graph_level_d_recursive(nrefine, b, v, x);
}

/****************************************************************************/

void  graph_level_d(GRAPH_WINDOW win, const DOF_REAL_D_VEC *v, REAL level,
		    const GRAPH_RGBCOLOR c, int refine)
{
  FUNCNAME("graph_level_d");
#if HAVE_LIBGL
  OGL_WINDOW *ogl_win = win;
#endif

  if (!v) return;

  if (level < 0.0) return;  /* no vector norms < 0 */

  if (v->fe_space && v->fe_space->admin && v->fe_space->admin->mesh)
  {
#if HAVE_LIBGL
    if (opengl_available)
    {
      OGL_set_std_window(ogl_win);
      glLineWidth(1.0);
      glColor3fv(c ? c : rgb_black);
    }
    else
#endif
    {
      PLOT_set_std_window((PLOT_INFO *)win);
      PLOT_set_linecolor((float *) (c ? c : rgb_black));
    }

    bas_fcts     = v->fe_space->bas_fcts;
    n_bas_fcts   = bas_fcts->n_bas_fcts;
    phi          = bas_fcts->phi;
    drdv         = v;
    level_value  = level;
    if (refine >= 0)
      nrefine = refine;
    else
      nrefine = MAX(0, bas_fcts->degree-1);

    mesh_traverse(v->fe_space->admin->mesh, -1, CALL_LEAF_EL | FILL_COORDS,
		  graph_level_d_fct);
    
#if HAVE_LIBGL
    if (opengl_available)
      OGL_FLUSH(ogl_win);
    else
#endif
    {
      PLOT_flush();
    }
  }
  else
    ERROR("no FE_SPACE OR DOF_ADMIN or MESH\n");
}

/****************************************************************************/
static REAL val_min, val_max;

static void val_minmax_fct(const EL_INFO *elinfo)
{
  FUNCNAME("val_minmax_fct");
#define NTEST 4
  static REAL b[NTEST][DIM+1] = {{1.0, 0.0, 0.0}, {0.0, 1.0, 0.0},
				 {0.0, 0.0, 1.0}, {0.333,0.333,0.334}};
  int  i, j, ntest;
  REAL v;

  if (bas_fcts->get_real_vec)
    el_vec = bas_fcts->get_real_vec(elinfo->el, drv, nil);
  else
    ERROR("no bas_fcts->get_real_vec()\n");

  if (nrefine>0) ntest = NTEST;
  else           ntest = 3;

  for (i = 0; i < ntest; i++)
  {
    v = 0.0;
    for (j=0; j<n_bas_fcts; j++)
      v += el_vec[j] * phi[j](b[i]);

    val_min = MIN(val_min, v);
    val_max = MAX(val_max, v);
  }
#undef NTEST
}

static void val_d_minmax_fct(const EL_INFO *elinfo)
{
  FUNCNAME("val_d_minmax_fct");
#define NTEST 4
  static REAL b[NTEST][DIM+1] = {{1.0, 0.0, 0.0}, {0.0, 1.0, 0.0},
				 {0.0, 0.0, 1.0}, {0.333,0.333,0.334}};
  int    i, j, k, ntest;
  REAL   v, phi_b;
  REAL_D vd;

  if (bas_fcts->get_real_d_vec)
    el_vec_d = bas_fcts->get_real_d_vec(elinfo->el, drdv, nil);
  else
    ERROR("no bas_fcts->get_real_d_vec()\n");

  if (nrefine>0) ntest = NTEST;
  else           ntest = 3;

  for (i = 0; i < ntest; i++)
  {
    SET_DOW(0.0, vd);
    for (j=0; j<n_bas_fcts; j++) {
	phi_b = phi[j](b[i]);
	for (k=0; k<DIM_OF_WORLD; ++k)
	    vd[k] += el_vec_d[j][k] * phi_b;
    }
    v = NORM_DOW(vd);

    val_min = MIN(val_min, v);
    val_max = MAX(val_max, v);
  }
#undef NTEST
}

void graph_levels(GRAPH_WINDOW win, const DOF_REAL_VEC *v,
		  int n, REAL const *levels, const GRAPH_RGBCOLOR *color,
		  int refine)
{
  FUNCNAME("graph_levels");
#define MAXLEV 100
  REAL                  lev[MAXLEV];
  GRAPH_RGBCOLOR        col[MAXLEV];
  const REAL            *l;
  const GRAPH_RGBCOLOR  *c;
  int                   nl, i;
  float                 nlr;

  if (!v) return;

  nl = MIN(n, MAXLEV);
  nlr = 1.0 / (float)(MAX(nl,1));

  if (v->fe_space && v->fe_space->admin && v->fe_space->admin->mesh)
  {
/*     if (v->fe_space->admin->mesh->parametric)  MSG("parametric mesh!\n"); */

    if (levels)
    {
      l = levels;
    }
    else
    {
      drv          = v;
      bas_fcts     = v->fe_space->bas_fcts;
      n_bas_fcts   = bas_fcts->n_bas_fcts;
      phi          = bas_fcts->phi;

      if (refine >= 0)
	nrefine = refine;
      else
	nrefine = MAX(0, bas_fcts->degree-1);

      val_max = -(val_min = 1.0E20);
      mesh_traverse(v->fe_space->admin->mesh, -1, CALL_LEAF_EL,
		    val_minmax_fct);
      val_max = MAX(val_max, val_min + 1.0E-10);

      for (i = 0; i < nl; i++)
	lev[i] = val_min + ((i-0.5) * nlr) * (val_max-val_min);
      l = lev;
    }

    if (color)
    {
      c = color;
    }
    else
    {
      c = (const GRAPH_RGBCOLOR *) col;
      for (i = 0; i < nl; i++)
      {
	col[i][0] = i * nlr;
	col[i][1] = 4.0 * (i * nlr) * (1.0 - i*nlr);
	col[i][2] = 1.0 - i * nlr;
      }
    }

    for (i = 0; i < nl; i++)
      graph_level(win, v, l[i], c[i], refine);
  }
  else
    ERROR("no FE_SPACE or DOF_ADMIN or MESH\n");
}


void graph_levels_d(GRAPH_WINDOW win, const DOF_REAL_D_VEC *v, int n, 
		    REAL const *levels, const GRAPH_RGBCOLOR *color, 
		    int refine)
{
  FUNCNAME("graph_levels_d");
#define MAXLEV 100
  REAL                  lev[MAXLEV];
  GRAPH_RGBCOLOR        col[MAXLEV];
  const GRAPH_RGBCOLOR  *c;
  const REAL            *l;
  int                   nl, i;
  REAL                  nlr;

  if (!v) return;

  nl = MIN(n, MAXLEV);
  nlr = 1.0 / (float)(MAX(nl,1));

  if (v->fe_space && v->fe_space->admin && v->fe_space->admin->mesh)
  {
/*     if (v->fe_space->admin->mesh->parametric)  MSG("parametric mesh!\n"); */

    if (levels)
    {
      l = levels;
    }
    else
    {
      drdv         = v;
      bas_fcts     = v->fe_space->bas_fcts;
      n_bas_fcts   = bas_fcts->n_bas_fcts;
      phi          = bas_fcts->phi;

      if (refine >= 0)
	nrefine = refine;
      else
	nrefine = MAX(0, bas_fcts->degree-1);

      val_max = -(val_min = 1.0E20);
      mesh_traverse(v->fe_space->admin->mesh, -1, CALL_LEAF_EL,
		    val_d_minmax_fct);
      val_max = MAX(val_max, val_min + 1.0E-10);

      for (i = 0; i < nl; i++)
	lev[i] = val_min + ((i-0.5) * nlr) * (val_max-val_min);
      l = lev;
    }

    if (color)
    {
      c = color;
    }
    else
    {
      c = (const GRAPH_RGBCOLOR *) col;
      for (i = 0; i < nl; i++)
      {
	col[i][0] = i * nlr;
	col[i][1] = 4.0 * (i * nlr) * (1.0 - i*nlr);
	col[i][2] = 1.0 - i * nlr;
      }
    }

    for (i = 0; i < nl; i++)
      graph_level_d(win, v, l[i], c[i], refine);
  }
  else
    ERROR("no FE_SPACE or DOF_ADMIN or MESH\n");
}

/****************************************************************************/
static REAL val_fac = 1.0;

static float *val_color(REAL v)
{
  static GRAPH_RGBCOLOR c;
  float  vv = (float) ((v - val_min) * val_fac);
  
  vv = MIN(1.0, vv);
  vv = MAX(0.0, vv);

  c[0] = vv;
  c[1] = 4 * vv * (1.0 - vv);
  c[2] = 1.0 - vv;

  return(c);
}


static void graph_value_recursive(int refine, const REAL *b[3],
				  REAL v[3], const REAL *x[3])
{
  int    i,j;

  if (refine > 0)       /* refine and call recursively */
  {
    const REAL *bnew[3], *xnew[3];
    REAL       bm[3][DIM+1], vnew[3], vm[3], xm[3][DIM_OF_WORLD];

    for (j = 0; j <= DIM; j++)
    {
      bm[0][j] = 0.5*(b[1][j] + b[2][j]);
      bm[1][j] = 0.5*(b[0][j] + b[2][j]);
      bm[2][j] = 0.5*(b[0][j] + b[1][j]);
    }

    for (i = 0; i < 3; i++)
    {
      for (vm[i] = j = 0; j < n_bas_fcts; j++)
	vm[i] += el_vec[j]*phi[j](bm[i]);
    }

    if (el_parametric)
    {
      el_parametric->coord_to_world(el_info, nil, N_EDGES,
				    (const REAL (*)[DIM+1]) bm, xm);
    }
    else
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xm[0][j] = 0.5*(x[1][j] + x[2][j]);
	xm[1][j] = 0.5*(x[0][j] + x[2][j]);
	xm[2][j] = 0.5*(x[0][j] + x[1][j]);
      }
    }

    bnew[0] = b[0]; bnew[1] = bm[2]; bnew[2] = bm[1]; 
    xnew[0] = x[0]; xnew[1] = xm[2]; xnew[2] = xm[1]; 
    vnew[0] = v[0]; vnew[1] = vm[2]; vnew[2] = vm[1]; 
    graph_value_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[1]; bnew[2] = bm[0]; 
    xnew[0] = xm[2]; xnew[1] = x[1]; xnew[2] = xm[0]; 
    vnew[0] = vm[2]; vnew[1] = v[1]; vnew[2] = vm[0]; 
    graph_value_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = b[2]; bnew[1] = bm[1]; bnew[2] = bm[0]; 
    xnew[0] = x[2]; xnew[1] = xm[1]; xnew[2] = xm[0]; 
    vnew[0] = v[2]; vnew[1] = vm[1]; vnew[2] = vm[0]; 
    graph_value_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[0]; bnew[1] = bm[1]; bnew[2] = bm[2]; 
    xnew[0] = xm[0]; xnew[1] = xm[1]; xnew[2] = xm[2]; 
    vnew[0] = vm[0]; vnew[1] = vm[1]; vnew[2] = vm[2]; 
    graph_value_recursive(refine-1, bnew, vnew, xnew);

    return;
  }

  /* draw triangle */

#if HAVE_LIBGL
  if (opengl_available)
  {
    glBegin(GL_TRIANGLES);
    for (i = 0; i < N_VERTICES; i++) {
      glColor3fv(val_color(v[i]));
      glVertex2dv(x[i]);
    }
    glEnd();
  }
  else
#endif
  {
    static float xy[N_VERTICES][DIM_OF_WORLD];
    REAL vv = (v[0]+v[1]+v[2])/3.0;
    float *vc = val_color(vv);
    PLOT_set_patchcolor(vc);
    for (i = 0; i < N_VERTICES; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
	xy[i][j] = x[i][j];
    PLOT_vtriangle2d(xy[0], xy[1], xy[2]);
  }
}


static void graph_value_fct(const EL_INFO *elinfo)
{
  FUNCNAME("graph_value_fct");
  static REAL bnew[3][DIM+1]={{1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,1.0}};
  static const REAL *b[3]={bnew[0],bnew[1],bnew[2]};
  static const REAL *x[3];
  static REAL_D coord[N_VERTICES];
  static REAL   v[3];
  int         i, j;
  PARAMETRIC  *parametric = elinfo->mesh->parametric;

  if (bas_fcts->get_real_vec)
    el_vec = bas_fcts->get_real_vec(elinfo->el, drv, nil);
  else
    ERROR("no bas_fcts->get_real_vec()\n");

  el_info       = elinfo;
  if (parametric && parametric->init_element(elinfo, parametric)) {
    el_parametric = parametric;
    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, coord);
    for (i = 0; i < 3; i++)
      x[i] = coord[i];
  }
  else {
    el_parametric = nil;
    for (i = 0; i < 3; i++)
      x[i] = elinfo->coord[i];
  }

  for (i = 0; i < 3; i++)
  {
    v[i] = 0.0;
    for (j=0; j<n_bas_fcts; j++)
      v[i] += el_vec[j] * phi[j](b[i]);
  }
      
  graph_value_recursive(nrefine, b, v, x);
}


void graph_values(GRAPH_WINDOW win, const DOF_REAL_VEC *v,
		  REAL min, REAL max, int refine)
{
  FUNCNAME("graph_values");
#if HAVE_LIBGL
  OGL_WINDOW *ogl_win = win;
#endif

  TEST_EXIT(v && v->fe_space && v->fe_space->admin)
    ("no vec or fe_space or admin\n");

/*   if (v->fe_space->admin->mesh->parametric) MSG("parametric mesh!\n"); */


  drv          = v;
  bas_fcts     = v->fe_space->bas_fcts;
  n_bas_fcts   = bas_fcts->n_bas_fcts;
  phi          = bas_fcts->phi;

  if (refine >= 0)
    nrefine = refine;
  else
    nrefine = MAX(0, bas_fcts->degree-1);

  if (min < max) {
      val_min = min;
      val_max = max;
  }
  else {
      val_max = -(val_min = 1.0E20);
      mesh_traverse(v->fe_space->admin->mesh, -1, CALL_LEAF_EL, val_minmax_fct);
      val_max = MAX(val_max, val_min + 1.0E-10);
  }
  val_fac = 1.0 / (val_max - val_min);

#if HAVE_LIBGL
  if (opengl_available)
  {
    OGL_set_std_window(ogl_win);
  }
  else
#endif
  {
    PLOT_set_std_window((PLOT_INFO *)win);
  }

  mesh_traverse(v->fe_space->admin->mesh, -1, CALL_LEAF_EL|FILL_COORDS,
		graph_value_fct);

#if HAVE_LIBGL
  if (opengl_available)
    OGL_FLUSH(ogl_win);
  else
#endif
    PLOT_flush();
}

/****************************************************************************/

static void graph_value_d_recursive(int refine, const REAL *b[3],
				  REAL v[3], const REAL *x[3])
{
  int    i, j, k;
  REAL   phi_b;

  if (refine > 0)       /* refine and call recursively */
  {
    const REAL *bnew[3], *xnew[3];
    REAL       bm[3][DIM+1], vnew[3], vm[3], xm[3][DIM_OF_WORLD];
    REAL_D     vd;

    for (j = 0; j <= DIM; j++)
    {
      bm[0][j] = 0.5*(b[1][j] + b[2][j]);
      bm[1][j] = 0.5*(b[0][j] + b[2][j]);
      bm[2][j] = 0.5*(b[0][j] + b[1][j]);
    }

    for (i = 0; i < 3; i++)
    {
      SET_DOW(0.0, vd);
      for (j = 0; j < n_bas_fcts; j++) 
      {
	phi_b = phi[j](bm[i]);
	for (k=0; k<DIM_OF_WORLD; k++)
	  vd[k] += el_vec_d[j][k]*phi_b;
      }
      vm[i] = NORM_DOW(vd);
    }

    if (el_parametric)
    {
      el_parametric->coord_to_world(el_info, nil, N_EDGES,
				    (const REAL (*)[DIM+1]) bm, xm);
    }
    else
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xm[0][j] = 0.5*(x[1][j] + x[2][j]);
	xm[1][j] = 0.5*(x[0][j] + x[2][j]);
	xm[2][j] = 0.5*(x[0][j] + x[1][j]);
      }
    }

    bnew[0] = b[0]; bnew[1] = bm[2]; bnew[2] = bm[1]; 
    xnew[0] = x[0]; xnew[1] = xm[2]; xnew[2] = xm[1]; 
    vnew[0] = v[0]; vnew[1] = vm[2]; vnew[2] = vm[1]; 
    graph_value_d_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[1]; bnew[2] = bm[0]; 
    xnew[0] = xm[2]; xnew[1] = x[1]; xnew[2] = xm[0]; 
    vnew[0] = vm[2]; vnew[1] = v[1]; vnew[2] = vm[0]; 
    graph_value_d_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = b[2]; bnew[1] = bm[1]; bnew[2] = bm[0]; 
    xnew[0] = x[2]; xnew[1] = xm[1]; xnew[2] = xm[0]; 
    vnew[0] = v[2]; vnew[1] = vm[1]; vnew[2] = vm[0]; 
    graph_value_d_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[0]; bnew[1] = bm[1]; bnew[2] = bm[2]; 
    xnew[0] = xm[0]; xnew[1] = xm[1]; xnew[2] = xm[2]; 
    vnew[0] = vm[0]; vnew[1] = vm[1]; vnew[2] = vm[2]; 
    graph_value_d_recursive(refine-1, bnew, vnew, xnew);

    return;
  }

  /* draw triangle */

#if HAVE_LIBGL
  if (opengl_available)
  {
    glBegin(GL_TRIANGLES);
    for (i = 0; i < N_VERTICES; i++) {
      glColor3fv(val_color(v[i]));
      glVertex2dv(x[i]);
    }
    glEnd();
  }
  else
#endif
  {
    static float xy[N_VERTICES][DIM_OF_WORLD];
    REAL vv = (v[0]+v[1]+v[2])/3.0;
    float *vc = val_color(vv);
    PLOT_set_patchcolor(vc);
    for (i = 0; i < N_VERTICES; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
	xy[i][j] = x[i][j];
    PLOT_vtriangle2d(xy[0], xy[1], xy[2]);
  }
}


static void graph_value_d_fct(const EL_INFO *elinfo)
{
  FUNCNAME("graph_value_d_fct");
  static REAL bnew[3][DIM+1]={{1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,1.0}};
  static const REAL *b[3]={bnew[0],bnew[1],bnew[2]};
  static const REAL *x[3];
  static REAL_D coord[N_VERTICES];
  static REAL   v[3], phi_b;
  static REAL_D vd;
  int         i, j, k;
  PARAMETRIC  *parametric = elinfo->mesh->parametric;

  if (bas_fcts->get_real_d_vec)
    el_vec_d = bas_fcts->get_real_d_vec(elinfo->el, drdv, nil);
  else
    ERROR("no bas_fcts->get_real_d_vec()\n");

  el_info       = elinfo;
  if (parametric && parametric->init_element(elinfo, parametric)) {
    el_parametric = parametric;
    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, coord);
    for (i = 0; i < 3; i++)
      x[i] = coord[i];
  }
  else {
    el_parametric = nil;
    for (i = 0; i < 3; i++)
      x[i] = elinfo->coord[i];
  }

  for (i = 0; i < 3; i++)
  {
    SET_DOW(0.0, vd);
    for (j = 0; j < n_bas_fcts; j++)
    {
      phi_b = phi[j](b[i]);
      for (k = 0; k < DIM_OF_WORLD; k++)
	vd[k] += el_vec_d[j][k]*phi_b;
    }
    v[i] = NORM_DOW(vd);
  }
      
  graph_value_d_recursive(nrefine, b, v, x);
}


void graph_values_d(GRAPH_WINDOW win, const DOF_REAL_D_VEC *v,
		    REAL min, REAL max, int refine)
{
  FUNCNAME("graph_values");
#if HAVE_LIBGL
  OGL_WINDOW *ogl_win = win;
#endif

  TEST_EXIT(v && v->fe_space && v->fe_space->admin)
    ("no vec or fe_space or admin\n");

/*   if (v->fe_space->admin->mesh->parametric) MSG("parametric mesh!\n"); */


  drdv         = v;
  bas_fcts     = v->fe_space->bas_fcts;
  n_bas_fcts   = bas_fcts->n_bas_fcts;
  phi          = bas_fcts->phi;

  if (refine >= 0)
    nrefine = refine;
  else
    nrefine = MAX(0, bas_fcts->degree-1);

  if (min < max) {
      val_min = min;
      val_max = max;
  }
  else {
      val_max = -(val_min = 1.0E20);
      mesh_traverse(v->fe_space->admin->mesh, -1, CALL_LEAF_EL, val_d_minmax_fct);
      val_max = MAX(val_max, val_min + 1.0E-10);
  }
  val_fac = 1.0 / (val_max - val_min);


#if HAVE_LIBGL
  if (opengl_available)
  {
    OGL_set_std_window(ogl_win);
  }
  else
#endif
  {
    PLOT_set_std_window((PLOT_INFO *)win);
  }

  mesh_traverse(v->fe_space->admin->mesh, -1, CALL_LEAF_EL|FILL_COORDS,
		graph_value_d_fct);

#if HAVE_LIBGL
  if (opengl_available)
    OGL_FLUSH(ogl_win);
  else
#endif
    PLOT_flush();
}


/****************************************************************************/
/*  graph_fvalue: general graph_value without DOF_REAL_VEC                  */
/****************************************************************************/
static REAL (*fvalue_fct)(const EL_INFO *, const REAL *lambda) = nil;

static void gval_minmax_fct(const EL_INFO *elinfo)
{
  FUNCNAME("gval_minmax_fct");
#define NTEST 4
  static REAL b[NTEST][DIM+1] = {{1.0, 0.0, 0.0}, {0.0, 1.0, 0.0},
				 {0.0, 0.0, 1.0}, {0.333,0.333,0.334}};
  int  i, ntest;
  REAL v;

  TEST_EXIT(fvalue_fct)("no fvalue_fct\n");

  if (nrefine>0) ntest = NTEST;
  else           ntest = 3;

  for (i = 0; i < ntest; i++)
  {
    v = fvalue_fct(elinfo, b[i]);

    val_min = MIN(val_min, v);
    val_max = MAX(val_max, v);
  }
#undef NTEST
}

/****************************************************************************/

static void graph_fvalue_recursive(int refine, const REAL *b[3],
				   REAL v[3], const REAL *x[3])
{
  int    i,j;

  if (refine > 0)       /* refine and call recursively */
  {
    const REAL *bnew[3], *xnew[3];
    REAL       bm[3][DIM+1], vnew[3], vm[3], xm[3][DIM_OF_WORLD];

    for (j = 0; j <= DIM; j++)
    {
      bm[0][j] = 0.5*(b[1][j] + b[2][j]);
      bm[1][j] = 0.5*(b[0][j] + b[2][j]);
      bm[2][j] = 0.5*(b[0][j] + b[1][j]);
    }

    for (i = 0; i < 3; i++)
    {
      vm[i] = fvalue_fct(el_info, bm[i]);
    }

    if (el_parametric)
    {
      el_parametric->coord_to_world(el_info, nil, N_EDGES,
				    (const REAL (*)[DIM+1]) bm, xm);
    }
    else
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xm[0][j] = 0.5*(x[1][j] + x[2][j]);
	xm[1][j] = 0.5*(x[0][j] + x[2][j]);
	xm[2][j] = 0.5*(x[0][j] + x[1][j]);
      }
    }

    bnew[0] = b[0]; bnew[1] = bm[2]; bnew[2] = bm[1]; 
    xnew[0] = x[0]; xnew[1] = xm[2]; xnew[2] = xm[1]; 
    vnew[0] = v[0]; vnew[1] = vm[2]; vnew[2] = vm[1]; 
    graph_fvalue_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[2]; bnew[2] = bm[1]; 
    xnew[0] = xm[2]; xnew[1] = x[2]; xnew[2] = xm[1]; 
    vnew[0] = vm[2]; vnew[1] = v[2]; vnew[2] = vm[1]; 
    graph_fvalue_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = b[2]; bnew[1] = bm[2]; bnew[2] = bm[0]; 
    xnew[0] = x[2]; xnew[1] = xm[2]; xnew[2] = xm[0]; 
    vnew[0] = v[2]; vnew[1] = vm[2]; vnew[2] = vm[0]; 
    graph_fvalue_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[1]; bnew[2] = bm[0]; 
    xnew[0] = xm[2]; xnew[1] = x[1]; xnew[2] = xm[0]; 
    vnew[0] = vm[2]; vnew[1] = v[1]; vnew[2] = vm[0]; 
    graph_fvalue_recursive(refine-1, bnew, vnew, xnew);

    return;
  }

  /* draw triangle */

#if HAVE_LIBGL
  if (opengl_available)
  {
    glBegin(GL_TRIANGLES);
    for (i = 0; i < N_VERTICES; i++) {
      glColor3fv(val_color(v[i]));
      glVertex2dv(x[i]);
    }
    glEnd();
  }
  else
#endif
  {
    static float xy[N_VERTICES][DIM_OF_WORLD];
    REAL vv = (v[0]+v[1]+v[2])/3.0;
    float *vc = val_color(vv);
    PLOT_set_patchcolor(vc);
    for (i = 0; i < N_VERTICES; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
	xy[i][j] = (float) x[i][j];
    PLOT_vtriangle2d(xy[0], xy[1], xy[2]);
  }
}


static void graph_fvalue_fct(const EL_INFO *elinfo)
{
  FUNCNAME("graph_fvalue_fct");
  static REAL   bnew[3][DIM+1]={{1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,1.0}};
  static const  REAL *b[3]={bnew[0],bnew[1],bnew[2]};
  static const  REAL *x[3];
  static REAL_D coord[N_VERTICES];
  static REAL   v[3];
  int           i;
  PARAMETRIC  *parametric = elinfo->mesh->parametric;

  TEST_EXIT(fvalue_fct)("no fvalue_fct\n");

  el_info       = elinfo;
  if (parametric && parametric->init_element(elinfo, parametric)) {
    el_parametric = parametric;
    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, coord);
    for (i = 0; i < 3; i++)
      x[i] = coord[i];
  }
  else {
    el_parametric = nil;
    for (i = 0; i < 3; i++)
      x[i] = elinfo->coord[i];
  }

  for (i = 0; i < 3; i++)
  {
    v[i] = fvalue_fct(elinfo, b[i]);
  }

  graph_fvalue_recursive(nrefine, b, v, x);
}

void graph_fvalues(GRAPH_WINDOW win, MESH *mesh,
		   REAL(*fct)(const EL_INFO *, const REAL *lambda),
		   FLAGS fill_flag, REAL min, REAL max, int refine)
{
  FUNCNAME("graph_fvalues");
#if HAVE_LIBGL
  OGL_WINDOW *ogl_win = win;
#endif

  TEST_EXIT(mesh)("no mesh\n");
  TEST_EXIT(fvalue_fct = fct)("no fct\n");

  nrefine = MAX(refine, 0);

  if (min < max) {
      val_min = min;
      val_max = max;
  }
  else {
      val_max = -(val_min = 1.0E20);
      mesh_traverse(mesh, -1, CALL_LEAF_EL | fill_flag, gval_minmax_fct);
      val_max = MAX(val_max, val_min + 1.0E-30);
  }
  val_fac = 1.0 / (val_max - val_min);


#if HAVE_LIBGL
  if (opengl_available)
    OGL_set_std_window(ogl_win);
  else
#endif
    PLOT_set_std_window((PLOT_INFO *)win);

  mesh_traverse(mesh, -1, CALL_LEAF_EL|FILL_COORDS, graph_fvalue_fct);

#if HAVE_LIBGL
  if (opengl_available)
    OGL_FLUSH(ogl_win);
  else
#endif
  {
    PLOT_flush();
  }
}

/****************************************************************************/
/* display of error estimators                                              */
/****************************************************************************/

static REAL (*Get_el_est)(EL *el) = nil;

static REAL el_est_fct(const EL_INFO *elinfo, const REAL *lambda)
{
  return(Get_el_est(elinfo->el));
}

static void est_minmax_fct(const EL_INFO *elinfo)
{
    REAL est = Get_el_est(elinfo->el);
    val_min = MIN(val_min, est);
    val_max = MAX(val_max, est);
}

void graph_el_est(GRAPH_WINDOW win, MESH *mesh, REAL (*get_el_est)(EL *el),
		  REAL min, REAL max)
{
  FUNCNAME("graph_el_est");

  TEST_EXIT(Get_el_est = get_el_est)("no get_el_est()\n");

  if (min >= max) {
      val_max = -(val_min = 1.0E20);
      mesh_traverse(mesh, -1, CALL_LEAF_EL, est_minmax_fct);
      min = val_min;
      max = val_max;
  }

  graph_fvalues(win, mesh, el_est_fct, 0, min, max, 0);
}

/****************************************************************************/
/****************************************************************************/
/* MULTIGRID LEVEL DISPLAYS                                                 */
/****************************************************************************/
/****************************************************************************/

void  graph_mesh_mg(GRAPH_WINDOW win, MESH *mesh, const GRAPH_RGBCOLOR c,
		    FLAGS flags, int mg_level)
{
  if (!mesh) return;

#if HAVE_LIBGL
  if (opengl_available) 
  {
    OGL_WINDOW *ogl_win = win;

    OGL_set_std_window(ogl_win);
    if (c) linecolor = c;
    else   linecolor = rgb_black;
    graph_mesh_flags = flags;
    mesh_traverse(mesh, mg_level, CALL_MG_LEVEL | FILL_COORDS | FILL_BOUND,
		  OGLgraph_mesh_fct);

    OGL_FLUSH(ogl_win);
    return;
  }
#endif

  PLOT_set_std_window((PLOT_INFO *)win);
  PLOT_set_linecolor((float *)  (c ? c : rgb_white));

  graph_mesh_flags = flags;
  mesh_traverse(mesh, mg_level, CALL_MG_LEVEL | FILL_COORDS | FILL_BOUND,
		Xgraph_mesh_fct);

  PLOT_flush();

}

/****************************************************************************/
static const int  *mg_sort_dof_invers = nil;
static int        mg_el_vec_size = 0;
static REAL       *mg_el_vec = nil;
static const REAL *drv_vec = nil;
static const DOF_ADMIN *mg_admin = nil;


static void val_minmax_mg_fct(const EL_INFO *elinfo)
{
  FUNCNAME("val_minmax_mg_fct");
#define NTEST 4
  static REAL b[NTEST][DIM+1] = {{1.0, 0.0, 0.0}, {0.0, 1.0, 0.0},
				 {0.0, 0.0, 1.0}, {0.333,0.333,0.334}};
  int       i, j, ntest;
  REAL      v;
  const DOF *dof_indices;

  if (bas_fcts->get_dof_indices)
    dof_indices = bas_fcts->get_dof_indices(elinfo->el, mg_admin, nil);
  else {
    ERROR("no bas_fcts->get_dof_indices()\n");
    return;
  }

  if (mg_sort_dof_invers) {
    for (i = 0; i < n_bas_fcts; i++) {
      mg_el_vec[i] = drv_vec[mg_sort_dof_invers[dof_indices[i]]];
    }
  }
  else  /* no mg_sort_dof_invers */
  {
    for (i = 0; i < n_bas_fcts; i++) {
      mg_el_vec[i] = drv_vec[dof_indices[i]];
    }
  }

  if (nrefine>0) ntest = NTEST;
  else           ntest = 3;

  for (i = 0; i < ntest; i++)
  {
    v = 0.0;
    for (j=0; j<n_bas_fcts; j++)
      v += mg_el_vec[j] * phi[j](b[i]);

    val_min = MIN(val_min, v);
    val_max = MAX(val_max, v);
  }
#undef NTEST
}



static void graph_value_mg_recursive(int refine, const REAL *b[3],
				  REAL v[3], const REAL *x[3])
{
  int    i, j;

  if (refine > 0)       /* refine and call recursively */
  {
    const REAL *bnew[3], *xnew[3];
    REAL       bm[3][DIM+1], vnew[3], vm[3], xm[3][DIM_OF_WORLD];

    for (j = 0; j <= DIM; j++)
    {
      bm[0][j] = 0.5*(b[1][j] + b[2][j]);
      bm[1][j] = 0.5*(b[0][j] + b[2][j]);
      bm[2][j] = 0.5*(b[0][j] + b[1][j]);
    }

    for (i = 0; i < 3; i++)
    {
      vm[i] = 0.0;
      for (j=0; j<n_bas_fcts; j++)
	vm[i] += mg_el_vec[j] * phi[j](bm[i]);
    }

    if (el_parametric)
    {
      el_parametric->coord_to_world(el_info, nil, N_EDGES,
				    (const REAL (*)[DIM+1]) bm, xm);
    }
    else
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xm[0][j] = 0.5*(x[1][j] + x[2][j]);
	xm[1][j] = 0.5*(x[0][j] + x[2][j]);
	xm[2][j] = 0.5*(x[0][j] + x[1][j]);
      }
    }

    bnew[0] = b[0]; bnew[1] = bm[2]; bnew[2] = bm[1]; 
    xnew[0] = x[0]; xnew[1] = xm[2]; xnew[2] = xm[1]; 
    vnew[0] = v[0]; vnew[1] = vm[2]; vnew[2] = vm[1]; 
    graph_value_mg_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[2]; bnew[2] = bm[1]; 
    xnew[0] = xm[2]; xnew[1] = x[2]; xnew[2] = xm[1]; 
    vnew[0] = vm[2]; vnew[1] = v[2]; vnew[2] = vm[1]; 
    graph_value_mg_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = b[2]; bnew[1] = bm[2]; bnew[2] = bm[0]; 
    xnew[0] = x[2]; xnew[1] = xm[2]; xnew[2] = xm[0]; 
    vnew[0] = v[2]; vnew[1] = vm[2]; vnew[2] = vm[0]; 
    graph_value_mg_recursive(refine-1, bnew, vnew, xnew);

    bnew[0] = bm[2]; bnew[1] = b[1]; bnew[2] = bm[0]; 
    xnew[0] = xm[2]; xnew[1] = x[1]; xnew[2] = xm[0]; 
    vnew[0] = vm[2]; vnew[1] = v[1]; vnew[2] = vm[0]; 
    graph_value_mg_recursive(refine-1, bnew, vnew, xnew);

    return;
  }

  /* draw triangle */

#if HAVE_LIBGL
  if (opengl_available)
  {
    glBegin(GL_TRIANGLES);
    for (i = 0; i < N_VERTICES; i++) {
      glColor3fv(val_color(v[i]));
      glVertex2dv(x[i]);
    }
    glEnd();
  }
  else
#endif
  {
    static float xy[N_VERTICES][DIM_OF_WORLD];
    REAL vv = (v[0]+v[1]+v[2])/3.0;
    float *vc = val_color(vv);
    PLOT_set_patchcolor(vc);
    for (i = 0; i < N_VERTICES; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
	xy[i][j] = (float) x[i][j];
    PLOT_vtriangle2d(xy[0], xy[1], xy[2]);
  }
}


static void graph_value_mg_fct(const EL_INFO *elinfo)
{
  FUNCNAME("graph_value_mg_fct");
  static REAL bnew[3][DIM+1]={{1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,1.0}};
  static const REAL *b[3]={bnew[0],bnew[1],bnew[2]};
  const REAL  *x[3];
  REAL_D      coord[N_VERTICES];
  REAL        v[3];
  int         i, j;
  const DOF   *dof_indices;
  PARAMETRIC  *parametric = elinfo->mesh->parametric;

  if (bas_fcts->get_dof_indices)
    dof_indices = bas_fcts->get_dof_indices(elinfo->el, mg_admin, nil);
  else {
    ERROR("no bas_fcts->get_dof_indices()\n");
    return;
  }

  if (mg_sort_dof_invers) {
    for (i = 0; i < n_bas_fcts; i++) {
      mg_el_vec[i] = drv_vec[mg_sort_dof_invers[dof_indices[i]]];
    }
  }
  else  /* no mg_sort_dof_invers */
  {
    for (i = 0; i < n_bas_fcts; i++) {
      mg_el_vec[i] = drv_vec[dof_indices[i]];
    }
  }

  el_info       = elinfo;
  if (parametric && parametric->init_element(elinfo, parametric)) {
    el_parametric = parametric;
    parametric->coord_to_world(elinfo, nil, N_VERTICES, vertex_lambda, coord);
    for (i = 0; i < 3; i++)
      x[i] = coord[i];
  }
  else {
    el_parametric = nil;
    for (i = 0; i < 3; i++)
      x[i] = elinfo->coord[i];
  }

  for (i = 0; i < 3; i++)
  {
    v[i] = 0.0;
    for (j=0; j<n_bas_fcts; j++)
      v[i] += mg_el_vec[j] * phi[j](b[i]);
  }

  graph_value_mg_recursive(nrefine, b, v, x);
}


void graph_values_mg(GRAPH_WINDOW win, const DOF_REAL_VEC *v,  
		     REAL min, REAL max, int refine,
		     int mg_level, const FE_SPACE *fe_space,
		     const int *sort_dof_invers)
{
  FUNCNAME("graph_values_mg");
#if HAVE_LIBGL
  OGL_WINDOW *ogl_win = win;
#endif

  TEST_EXIT(v && fe_space && fe_space->admin)
    ("no vec or fe_space or admin\n");

  drv          = v;
  drv_vec      = drv->vec;
  bas_fcts     = fe_space->bas_fcts;
  mg_admin     = fe_space->admin;
  n_bas_fcts   = bas_fcts->n_bas_fcts;
  phi          = bas_fcts->phi;

  mg_sort_dof_invers = sort_dof_invers;
  if (n_bas_fcts > mg_el_vec_size) {
    mg_el_vec = MEM_REALLOC(mg_el_vec, mg_el_vec_size, n_bas_fcts, REAL);
    mg_el_vec_size = n_bas_fcts;
  }

  if (refine >= 0)
    nrefine = refine;
  else
    nrefine = MAX(0, bas_fcts->degree-1);

  
  if (min < max) {
      val_min = min;
      val_max = max;
  }
  else {
      val_max = -(val_min = 1.0E20);
      mesh_traverse(fe_space->admin->mesh, mg_level,
		    CALL_MG_LEVEL, val_minmax_mg_fct);
/*       MSG("<%s> value range in [%.3le , %.3le]\n", NAME(drv), val_min, val_max); */

      val_max = MAX(val_max, val_min + 1.0E-5);
  }
  val_fac = 1.0 / (val_max - val_min);


#if HAVE_LIBGL
  if (opengl_available)
  {
    OGL_set_std_window(ogl_win);
  }
  else
#endif
  {
    PLOT_set_std_window((PLOT_INFO *)win);
  }

  mesh_traverse(fe_space->admin->mesh, mg_level, CALL_MG_LEVEL | FILL_COORDS,
		graph_value_mg_fct);



  graph_mesh_mg(win, fe_space->admin->mesh, rgb_black, 0, mg_level);


#if HAVE_LIBGL
  if (opengl_available)
    OGL_FLUSH(ogl_win);
  else
#endif
  {
    PLOT_flush();
  }
}

/****************************************************************************/

void graph_drv(GRAPH_WINDOW win, const DOF_REAL_VEC *u,
	       REAL min, REAL max, int refine)
{
  graph_values(win, u, min, max, refine);
}

void graph_drv_d(GRAPH_WINDOW win, const DOF_REAL_D_VEC *ud,
		 REAL min, REAL max, int refine)
{
  graph_values_d(win, ud, min, max, refine);
}
