/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     lagrange_3.c                                                   */
/*                                                                          */
/* description:  piecewise cubic Lagrange elements in 3d                    */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#define N_BAS3  20

static const REAL   bary3[N_BAS3][DIM+1] = {{1.0, 0.0, 0.0, 0.0},
					    {0.0, 1.0, 0.0, 0.0},
					    {0.0, 0.0, 1.0, 0.0},
					    {0.0, 0.0, 0.0, 1.0},
					    {2.0/3.0,1.0/3.0,    0.0,    0.0},
					    {1.0/3.0,2.0/3.0,    0.0,    0.0},
					    {2.0/3.0,    0.0,1.0/3.0,    0.0},
					    {1.0/3.0,    0.0,2.0/3.0,    0.0},
					    {2.0/3.0,    0.0,    0.0,1.0/3.0},
					    {1.0/3.0,    0.0,    0.0,2.0/3.0},
					    {    0.0,2.0/3.0,1.0/3.0,    0.0},
					    {    0.0,1.0/3.0,2.0/3.0,    0.0},
					    {    0.0,2.0/3.0,    0.0,1.0/3.0},
					    {    0.0,1.0/3.0,    0.0,2.0/3.0},
					    {    0.0,    0.0,2.0/3.0,1.0/3.0},
					    {    0.0,    0.0,1.0/3.0,2.0/3.0},
					    {    0.0,1.0/3.0,1.0/3.0,1.0/3.0},
					    {1.0/3.0,    0.0,1.0/3.0,1.0/3.0},
					    {1.0/3.0,1.0/3.0,    0.0,1.0/3.0},
					    {1.0/3.0,1.0/3.0,1.0/3.0,    0.0}};
#if 0
static const int    fct_at_node3[N_BAS3] = {0,1,2,3,4,4,5,5,6,6,7,7,8,8,9,9,
					    10,11,12,13};

static const int    n_dof3[DIM+1] = {1, 2, 1, 0};
#endif

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3v0(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[0] - 1.0)*lambda[0] + 1.0)*lambda[0]);
}

static const REAL *grd_phi3v0(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = (13.5*lambda[0] - 9.0)*lambda[0] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3v0(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0},{0, 0, 0, 0}, {0, 0, 0, 0}};
  D2[0][0] = 27.0*lambda[0] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3v1(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[1] - 1.0)*lambda[1] + 1.0)*lambda[1]);
}

static const REAL *grd_phi3v1(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = (13.5*lambda[1] - 9.0)*lambda[1] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3v1(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0},{0, 0, 0, 0}, {0, 0, 0, 0}};
  D2[1][1] = 27.0*lambda[1] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3v2(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[2] - 1.0)*lambda[2] + 1.0)*lambda[2]);
}

static const REAL *grd_phi3v2(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[2] = (13.5*lambda[2] - 9.0)*lambda[2] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3v2(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0},{0, 0, 0, 0}, {0, 0, 0, 0}};
  D2[2][2] = 27.0*lambda[2] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3v3(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[3] - 1.0)*lambda[3] + 1.0)*lambda[3]);
}

static const REAL *grd_phi3v3(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[3] = (13.5*lambda[3] - 9.0)*lambda[3] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3v3(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0, 0},{0, 0, 0, 0}, {0, 0, 0, 0}};
  D2[3][3] = 27.0*lambda[3] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 0 inbetween vertex 0 and 1                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e0(const REAL lambda[DIM+1])
{
  return((13.5*lambda[0] - 4.5)*lambda[0]*lambda[1]);
}

static const REAL *grd_phi3e0(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = (27.0*lambda[0] - 4.5)*lambda[1];
  grd[1] = (13.5*lambda[0] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e0(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[0][0] = 27.0*lambda[1];
  D2[0][1] = D2[1][0] = 27.0*lambda[0] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 0 inbetween vertex 0 and 1                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e1(const REAL lambda[DIM+1])
{
  return((13.5*lambda[1] - 4.5)*lambda[1]*lambda[0]);
}

static const REAL *grd_phi3e1(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = (13.5*lambda[1] - 4.5)*lambda[1];
  grd[1] = (27.0*lambda[1] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e1(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[1][1] = 27.0*lambda[0];
  D2[0][1] = D2[1][0] = 27.0*lambda[1] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 1 inbetween vertex 0 and 2                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e2(const REAL lambda[DIM+1])
{
  return((13.5*lambda[0] - 4.5)*lambda[0]*lambda[2]);
}

static const REAL *grd_phi3e2(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = (27.0*lambda[0] - 4.5)*lambda[2];
  grd[2] = (13.5*lambda[0] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e2(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[0][0] = 27.0*lambda[2];
  D2[0][2] = D2[2][0] = 27.0*lambda[0] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 1 inbetween vertex 0 and 2                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e3(const REAL lambda[DIM+1])
{
  return((13.5*lambda[2] - 4.5)*lambda[2]*lambda[0]);
}

static const REAL *grd_phi3e3(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = (13.5*lambda[2] - 4.5)*lambda[2];
  grd[2] = (27.0*lambda[2] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e3(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[2][2] = 27.0*lambda[0];
  D2[0][2] = D2[2][0] = 27.0*lambda[2] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 2 inbetween vertex 0 and 3                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e4(const REAL lambda[DIM+1])
{
  return((13.5*lambda[0] - 4.5)*lambda[0]*lambda[3]);
}

static const REAL *grd_phi3e4(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = (27.0*lambda[0] - 4.5)*lambda[3];
  grd[3] = (13.5*lambda[0] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e4(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[0][0] = 27.0*lambda[3];
  D2[0][3] = D2[3][0] = 27.0*lambda[0] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 0 inbetween vertex 0 and 3                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e5(const REAL lambda[DIM+1])
{
  return((13.5*lambda[3] - 4.5)*lambda[3]*lambda[0]);
}

static const REAL *grd_phi3e5(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = (13.5*lambda[3] - 4.5)*lambda[3];
  grd[3] = (27.0*lambda[3] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e5(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[3][3] = 27.0*lambda[0];
  D2[0][3] = D2[3][0] = 27.0*lambda[3] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 3 inbetween vertex 1 and 2                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e6(const REAL lambda[DIM+1])
{
  return((13.5*lambda[1] - 4.5)*lambda[1]*lambda[2]);
}

static const REAL *grd_phi3e6(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = (27.0*lambda[1] - 4.5)*lambda[2];
  grd[2] = (13.5*lambda[1] - 4.5)*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e6(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[1][1] = 27.0*lambda[2];
  D2[1][2] = D2[2][1] = 27.0*lambda[1] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 3 inbetween vertex 1 and 2                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e7(const REAL lambda[DIM+1])
{
  return((13.5*lambda[2] - 4.5)*lambda[2]*lambda[1]);
}

static const REAL *grd_phi3e7(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = (13.5*lambda[2] - 4.5)*lambda[2];
  grd[2] = (27.0*lambda[2] - 4.5)*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e7(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[2][2] = 27.0*lambda[1];
  D2[1][2] = D2[2][1] = 27.0*lambda[2] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 4 inbetween vertex 1 and 3                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e8(const REAL lambda[DIM+1])
{
  return((13.5*lambda[1] - 4.5)*lambda[1]*lambda[3]);
}

static const REAL *grd_phi3e8(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = (27.0*lambda[1] - 4.5)*lambda[3];
  grd[3] = (13.5*lambda[1] - 4.5)*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e8(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[1][1] = 27.0*lambda[3];
  D2[1][3] = D2[3][1] = 27.0*lambda[1] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 4 inbetween vertex 1 and 3                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e9(const REAL lambda[DIM+1])
{
  return((13.5*lambda[3] - 4.5)*lambda[3]*lambda[1]);
}

static const REAL *grd_phi3e9(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = (13.5*lambda[3] - 4.5)*lambda[3];
  grd[3] = (27.0*lambda[3] - 4.5)*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e9(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[3][3] = 27.0*lambda[1];
  D2[1][3] = D2[3][1] = 27.0*lambda[3] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 5 inbetween vertex 2 and 3                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e10(const REAL lambda[DIM+1])
{
  return((13.5*lambda[2] - 4.5)*lambda[2]*lambda[3]);
}

static const REAL *grd_phi3e10(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[2] = (27.0*lambda[2] - 4.5)*lambda[3];
  grd[3] = (13.5*lambda[2] - 4.5)*lambda[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e10(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[2][2] = 27.0*lambda[3];
  D2[2][3] = D2[3][2] = 27.0*lambda[2] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 5 inbetween vertex 2 and 3                      */
/*--------------------------------------------------------------------------*/

static REAL phi3e11(const REAL lambda[DIM+1])
{
  return((13.5*lambda[3] - 4.5)*lambda[3]*lambda[2]);
}

static const REAL *grd_phi3e11(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[2] = (13.5*lambda[3] - 4.5)*lambda[3];
  grd[3] = (27.0*lambda[3] - 4.5)*lambda[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e11(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[3][3] = 27.0*lambda[2];
  D2[2][3] = D2[3][2] = 27.0*lambda[3] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at barycenter of face 0                                   */
/*--------------------------------------------------------------------------*/

static REAL phi3c0(const REAL lambda[DIM+1])
{
  return(27.0*lambda[1]*lambda[2]*lambda[3]);
}

static const REAL *grd_phi3c0(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[1] = 27.0*lambda[2]*lambda[3];
  grd[2] = 27.0*lambda[1]*lambda[3];
  grd[3] = 27.0*lambda[1]*lambda[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3c0(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[1][2] = D2[2][1] = 27.0*lambda[3];
  D2[1][3] = D2[3][1] = 27.0*lambda[2];
  D2[2][3] = D2[3][2] = 27.0*lambda[1];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at barycenter of face 1                                   */
/*--------------------------------------------------------------------------*/

static REAL phi3c1(const REAL lambda[DIM+1])
{
  return(27.0*lambda[0]*lambda[2]*lambda[3]);
}

static const REAL *grd_phi3c1(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = 27.0*lambda[2]*lambda[3];
  grd[2] = 27.0*lambda[0]*lambda[3];
  grd[3] = 27.0*lambda[0]*lambda[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3c1(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[0][2] = D2[2][0] = 27.0*lambda[3];
  D2[0][3] = D2[3][0] = 27.0*lambda[2];
  D2[2][3] = D2[3][2] = 27.0*lambda[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at barycenter of face 2                                   */
/*--------------------------------------------------------------------------*/

static REAL phi3c2(const REAL lambda[DIM+1])
{
  return(27.0*lambda[0]*lambda[1]*lambda[3]);
}

static const REAL *grd_phi3c2(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = 27.0*lambda[1]*lambda[3];
  grd[1] = 27.0*lambda[0]*lambda[3];
  grd[3] = 27.0*lambda[0]*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3c2(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[0][1] = D2[1][0] = 27.0*lambda[3];
  D2[0][3] = D2[3][0] = 27.0*lambda[1];
  D2[1][3] = D2[3][1] = 27.0*lambda[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at barycenter of face 3                                   */
/*--------------------------------------------------------------------------*/

static REAL phi3c3(const REAL lambda[DIM+1])
{
  return(27.0*lambda[0]*lambda[1]*lambda[2]);
}

static const REAL *grd_phi3c3(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0, 0.0};
  grd[0] = 27.0*lambda[1]*lambda[2];
  grd[1] = 27.0*lambda[0]*lambda[2];
  grd[2] = 27.0*lambda[0]*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3c3(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}, 
				   {0.0,0.0,0.0,0.0}, {0.0,0.0,0.0,0.0}};
  D2[0][1] = D2[1][0] = 27.0*lambda[2];
  D2[0][2] = D2[2][0] = 27.0*lambda[1];
  D2[1][2] = D2[2][1] = 27.0*lambda[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  functions for combining basisfunctions with coefficients                */
/*--------------------------------------------------------------------------*/

static const DOF *get_dof_indices3(const EL *el, const DOF_ADMIN *admin,
				   DOF *idof)
{
  static DOF  index_vec[N_BAS3];
  DOF         *rvec = idof ? idof : index_vec;
  int         i, j, n0;
  DOF         **dof = el->dof;

  n0 = admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = dof[i][n0];
  n0 = admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = dof[N_VERTICES+i][n0];
      rvec[j++] = dof[N_VERTICES+i][n0+1];
    }
    else
    {
      rvec[j++] = dof[N_VERTICES+i][n0+1];
      rvec[j++] = dof[N_VERTICES+i][n0];
    }
  }
  n0 = admin->n0_dof[FACE];
  for (i = 0; i < N_FACES; i++)
    rvec[j++] = dof[N_VERTICES+N_EDGES+i][n0];

  return((const DOF *) rvec);
}

static const S_CHAR *get_bound3(const EL_INFO *el_info, S_CHAR *bound)
{
  FUNCNAME("get_bound3");
  static S_CHAR  bound_vec[N_BAS3];
  S_CHAR         *rvec = bound ? bound : bound_vec;
  int            i, j;

  TEST_FLAG(FILL_BOUND, el_info);

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = el_info->bound[i];
  for (j = N_VERTICES, i = N_FACES; i < N_FACES+N_EDGES; j += 2, i++)
    rvec[j] = rvec[j+1] = GET_BOUND(el_info->boundary[i]);
  for (i = 0; i < N_FACES; i++)
    rvec[j++] = GET_BOUND(el_info->boundary[i]);
  
  return((const S_CHAR *) rvec);
}

static const int *get_int_vec3(const EL *el, const DOF_INT_VEC *vec,
				int *ivec)
{
  FUNCNAME("get_int_vec3");
  int         i, j, n0;
  static int  local_vec[N_BAS3];
  int         *v = nil, *rvec = ivec ? ivec : local_vec;
  DOF         **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
    }
    else
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
    }
  }
  n0 = vec->fe_space->admin->n0_dof[FACE];
  for (i = 0; i < N_FACES; i++)
    rvec[j++] = v[dof[N_VERTICES+N_EDGES+i][n0]];

  return((const int *) rvec);
}

static const REAL *get_real_vec3(const EL *el, const DOF_REAL_VEC *vec,
				  REAL *Rvec)
{
  FUNCNAME("get_real_vec3");
  int            i, j, n0;
  static REAL    local_vec[N_BAS3];
  REAL           *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
    }
    else
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
    }
  }
  n0 = vec->fe_space->admin->n0_dof[FACE];
  for (i = 0; i < N_FACES; i++)
    rvec[j++] = v[dof[N_VERTICES+N_EDGES+i][n0]];

  return((const REAL *) rvec);
}

static const REAL_D *get_real_d_vec3(const EL *el, const DOF_REAL_D_VEC *vec,
				     REAL_D *Rvec)
{
  FUNCNAME("get_real_d_vec3");
  int            i, j, k, n0;
  static REAL_D  local_vec[N_BAS3];
  REAL_D         *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    for (k = 0; k < DIM_OF_WORLD; k++)
      rvec[i][k] = v[dof[i][n0]][k];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	rvec[j][k] = v[dof[N_VERTICES+i][n0]][k];
	rvec[j+1][k] = v[dof[N_VERTICES+i][n0+1]][k];
      }
      j += 2;
    }
    else
    {
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	rvec[j][k] = v[dof[N_VERTICES+i][n0+1]][k];
	rvec[j+1][k] = v[dof[N_VERTICES+i][n0]][k];
      }
      j += 2;
    }
  }
  n0 = vec->fe_space->admin->n0_dof[FACE];
  for (i = 0; i < N_FACES; i++, j++)
    for (k = 0; k < DIM_OF_WORLD; k++)
      rvec[j][k] = v[dof[N_VERTICES+N_EDGES+i][n0]][k];

  return((const REAL_D *) rvec);
}

static const U_CHAR *get_uchar_vec3(const EL *el, const DOF_UCHAR_VEC *vec,
				     U_CHAR *uvec)
{
  FUNCNAME("get_uchar_vec3");
  int            i, j, n0;
  static U_CHAR  local_vec[N_BAS3];
  U_CHAR         *v = nil, *rvec = uvec ? uvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
    }
    else
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
    }
  }
  n0 = vec->fe_space->admin->n0_dof[FACE];
  for (i = 0; i < N_FACES; i++)
    rvec[j++] = v[dof[N_VERTICES+N_EDGES+i][n0]];

  return((const U_CHAR *) rvec);
}

static const S_CHAR *get_schar_vec3(const EL *el, const DOF_SCHAR_VEC *vec,
				     S_CHAR *svec)
{
  FUNCNAME("get_schar_vec3");
  int            i, j, n0;
  static S_CHAR  local_vec[N_BAS3];
  S_CHAR         *v = nil, *rvec = svec ? svec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
    }
    else
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
    }
  }
  n0 = vec->fe_space->admin->n0_dof[FACE];
  for (i = 0; i < N_FACES; i++)
    rvec[j++] = v[dof[N_VERTICES+N_EDGES+i][n0]];

  return((const S_CHAR *) rvec);
}

#if 0
static REAL  one_third = 1.0/3.0, two_third = 2.0/3.0;

static const REAL *bary_center_face(int face, const REAL_D *coord, REAL_D x)
{
  FUNCNAME("bary_center_face");
  int      j;
  REAL_D   bary;
  REAL     *vec = x ? x : bary;

  switch (face)
  {
  case 0:
    for (j = 0; j < DIM_OF_WORLD; j++)
      vec[j] = one_third*(coord[1][j] + coord[2][j] + coord[3][j]);
    return((const REAL *) vec);
  case 1:
    for (j = 0; j < DIM_OF_WORLD; j++)
      vec[j] = one_third*(coord[2][j] + coord[3][j] + coord[0][j]);
    return((const REAL *) vec);
  case 2:
    for (j = 0; j < DIM_OF_WORLD; j++)
      vec[j] = one_third*(coord[3][j] + coord[0][j] + coord[1][j]);
    return((const REAL *) vec);
  case 3:
    for (j = 0; j < DIM_OF_WORLD; j++)
      vec[j] = one_third*(coord[0][j] + coord[1][j] + coord[2][j]);
    return((const REAL *) vec);
  }
}  
#endif

static const REAL *interpol3(const EL_INFO *el_info, int no, 
			      const int *b_no, REAL (*f)(const REAL_D),
			      REAL *vec)
{
  FUNCNAME("interpol3");
  static REAL  inter[N_BAS3];
  REAL         *rvec = vec ? vec : inter;
  REAL_D       x;
  const REAL_D *coord;
  int          i;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (parametric) 
    ERROR_EXIT("Sorry, parametric support for lagrange3 not implemented yet!\n");

  TEST_FLAG(FILL_COORDS, el_info);
  coord = el_info->coord;

  if (b_no)
  {
    if (no <= 0  ||  no > N_BAS3)
    {
      MSG("something is wrong, doing nothing\n");
      rvec[0] = 0.0;
      return((const REAL *) rvec);
    }
  
    for (i = 0; i < no; i++)
    {
      if (b_no[i] < N_VERTICES)
      {
	rvec[i] = f(coord[b_no[i]]);
      }
      else 
      {
	coord_to_world(el_info, bary3[b_no[i]], x);
	rvec[i] = f(x);
      }
    }
  }
  else
  {
    for (i = 0; i < N_VERTICES; i++)
      rvec[i] = f(coord[i]);
    for (i = N_VERTICES; i < N_BAS3; i++)
    {
      coord_to_world(el_info, bary3[i], x);
      rvec[i] = f(x);
    }
  }  

  return((const REAL *) rvec);
}

static const REAL_D *interpol_d3(const EL_INFO *el_info, int no, 
				 const int *b_no,
				 const REAL *(*f)(const REAL_D, REAL_D),
				 REAL_D *vec)
{
  FUNCNAME("interpol_d3");
  static REAL_D  inter[N_BAS3];
  REAL_D         *rvec = vec ? vec : inter;
  REAL_D         x;
  const REAL_D   *coord;
  int            i;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (parametric) 
    ERROR_EXIT("Sorry, parametric support for lagrange3 not implemented yet!\n");

  TEST_FLAG(FILL_COORDS, el_info);
  coord = el_info->coord;

  if (b_no)
  {
    if (no <= 0  ||  no > N_BAS3)
    {
      ERROR("something is wrong, doing nothing\n");
      for (i = 0; i < DIM_OF_WORLD; i++)
	rvec[0][i] = 0.0;
      return((const REAL_D *) rvec);
    }
  
    for (i = 0; i < no; i++)
    {
      if (b_no[i] < N_VERTICES)
      {
	(*f)(coord[b_no[i]], rvec[i]);
      }
      else 
      {
	coord_to_world(el_info, bary3[b_no[i]], x);
	(*f)(x, rvec[i]);
      }
    }
  }
  else
  {
    for (i = 0; i < N_VERTICES; i++)
      (*f)(coord[i], rvec[i]);
    for (i = N_VERTICES; i < N_BAS3; i++)
    {
      coord_to_world(el_info, bary3[i], x);
      (*f)(x, rvec[i]);
    }
  }  

  return((const REAL_D *) rvec);
}

/*--------------------------------------------------------------------------*/
/*  functions for interpolation/ restriction during refinement/coarsening   */
/*--------------------------------------------------------------------------*/

static void real_refine_inter3(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_refine_inter3");
  EL        *el;
  REAL      *v = nil;
  const DOF *cd;
  DOF       pd[N_BAS3], cdi;
  int       i, typ, lr_set, node0, n0;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  v[cd[3]] = (0.0625*(-v[pd[0]] - v[pd[1]]) + 0.5625*(v[pd[4]] + v[pd[5]]));
  v[cd[8]] = (0.3125*(v[pd[0]] - v[pd[5]]) + 0.0625*v[pd[1]]
	      + 0.937500*v[pd[4]]);
  v[cd[9]] = v[pd[4]];
  v[cd[12]] = (0.0625*(v[pd[0]] + v[pd[1]] - v[pd[4]] - v[pd[5]])
	       + 0.25*(-v[pd[6]] - v[pd[10]])
	       + 0.5*(v[pd[7]] + v[pd[11]] + v[pd[19]]));
  v[cd[13]] = v[pd[19]];
  v[cd[14]] = (0.0625*(v[pd[0]] + v[pd[1]] - v[pd[4]] - v[pd[5]])
	       + 0.25*(-v[pd[8]] - v[pd[12]]) 
	       + 0.5*(v[pd[9]] + v[pd[13]] + v[pd[18]]));
  v[cd[15]] = (v[pd[18]]);
  v[cd[16]] = (0.0625*(v[pd[0]] + v[pd[1]] - v[pd[4]] - v[pd[5]])
	       + 0.125*(-v[pd[6]]-v[pd[8]]-v[pd[10]]-v[pd[12]])
	       + 0.5*(v[pd[16]] + v[pd[17]]) 
	       + 0.25*(v[pd[18]] + v[pd[19]]));
  v[cd[17]] = (0.0625*(-v[pd[0]] + v[pd[1]]) 
	       + 0.1875*(v[pd[4]] - v[pd[5]]) + 0.375*v[pd[8]]
	       - 0.125*v[pd[12]] + 0.75*v[pd[18]]);
  v[cd[18]] = (0.0625*(-v[pd[0]] + v[pd[1]]) 
	       + 0.1875*(v[pd[4]] - v[pd[5]]) + 0.375*v[pd[6]]
	       - 0.125*v[pd[10]] + 0.75*v[pd[19]]);

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);

  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    v[cd[8]] = (0.0625*v[pd[0]] + 0.3125*(v[pd[1]] - v[pd[4]])
		+ 0.9375*v[pd[5]]);
    v[cd[9]] = v[pd[5]];
    v[cd[17]] = (0.0625*(v[pd[0]] - v[pd[1]]) 
		 + 0.1875*(-v[pd[4]] + v[pd[5]])
		 - 0.125*v[pd[6]] + 0.375*v[pd[10]] + 0.75*v[pd[19]]);
    v[cd[18]] = (0.0625*(v[pd[0]] - v[pd[1]]) 
		 + 0.1875*(-v[pd[4]] + v[pd[5]]) - 0.125*v[pd[8]] 
		 + 0.375*v[pd[12]] + 0.75*v[pd[18]]);
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    v[cd[8]] = (0.0625*v[pd[0]] + 0.3125*(v[pd[1]] - v[pd[4]])
		+ 0.9375*v[pd[5]]);
    v[cd[9]] = v[pd[5]];
    v[cd[17]] = (0.0625*(v[pd[0]] - v[pd[1]]) 
		 + 0.1875*(-v[pd[4]] + v[pd[5]]) - 0.125*v[pd[8]]
		 + 0.375*v[pd[12]] + 0.75*v[pd[18]]);
    v[cd[18]] = (0.0625*(v[pd[0]] - v[pd[1]]) + 
		 0.1875*(-v[pd[4]] + v[pd[5]]) - 0.125*v[pd[6]]
		 + 0.375*v[pd[10]] + 0.75*v[pd[19]]);
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/

  node0 = drv->fe_space->mesh->node[FACE];
  n0 = admin->n0_dof[FACE];
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      v[cd[12]] = (0.0625*(v[pd[0]]+v[pd[1]]-v[pd[4]]-v[pd[5]])
		   + 0.25*(-v[pd[6]] - v[pd[10]]) 
		   + 0.5*(v[pd[7]] + v[pd[11]] + v[pd[19]]));
      v[cd[13]] = v[pd[19]];
      v[cd[16]] = (0.0625*(v[pd[0]]+v[pd[1]]-v[pd[4]]-v[pd[5]])
		   + 0.125*(-v[pd[6]]-v[pd[8]]-v[pd[10]]-v[pd[12]])
		   + 0.5*(v[pd[16]] + v[pd[17]]) 
		   + 0.25*(v[pd[18]] + v[pd[19]]));
      v[cd[18]] = (0.0625*(-v[pd[0]] + v[pd[1]])
		   + 0.1875*(v[pd[4]] - v[pd[5]]) + 0.375*v[pd[6]]
		   - 0.125*v[pd[10]] + 0.75*v[pd[19]]);
      break;
    case 2:
      v[cd[14]] = (0.0625*(v[pd[0]]+v[pd[1]]-v[pd[4]]-v[pd[5]])
		   + 0.25*(-v[pd[8]] - v[pd[12]]) 
		   + 0.5*(v[pd[9]] + v[pd[13]] + v[pd[18]]));
      v[cd[15]] = v[pd[18]];
      v[cd[16]] = (0.0625*(v[pd[0]]+v[pd[1]]-v[pd[4]]-v[pd[5]])
		   + 0.125*(-v[pd[6]]-v[pd[8]]-v[pd[10]]-v[pd[12]])
		   + 0.5*(v[pd[16]] + v[pd[17]])
		   + 0.25*(v[pd[18]] + v[pd[19]]));
      v[cd[17]] = (0.0625*(-v[pd[0]] + v[pd[1]])
		   + 0.1875*(v[pd[4]] - v[pd[5]]) + 0.375*v[pd[8]]
		   - 0.125*v[pd[12]] + 0.75*v[pd[18]]);
      break;
   case 3:
      v[cd[16]] = (0.0625*(v[pd[0]]+v[pd[1]]-v[pd[4]]-v[pd[5]])
		   + 0.125*(-v[pd[6]]-v[pd[8]]-v[pd[10]]-v[pd[12]])
		   + 0.5*(v[pd[16]] + v[pd[17]])
		   + 0.25*(v[pd[18]] + v[pd[19]]));
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	cdi = el->child[1]->dof[node0+1][n0];      
	TEST_EXIT(cdi == cd[17])("cdi != cd[17]\n");
	v[cdi] = (0.0625*(v[pd[0]] - v[pd[1]])
		  + 0.1875*(-v[pd[4]] + v[pd[5]]) - 0.125*v[pd[6]]
		  + 0.375*v[pd[10]] + 0.75*v[pd[19]]);
	break;
      case 2:
	cdi = el->child[1]->dof[node0+2][n0];      
	TEST_EXIT(cdi == cd[18])("cdi != cd[18]\n");
	v[cdi] = (0.0625*(v[pd[0]] - v[pd[1]])
		  + 0.1875*(-v[pd[4]] + v[pd[5]]) - 0.125*v[pd[8]]
		  + 0.375*v[pd[12]] + 0.75*v[pd[18]]);
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	cdi = el->child[1]->dof[node0+2][n0];      
	TEST_EXIT(cdi == cd[18])("cdi != cd[18]\n");
	v[cdi] = (0.0625*(v[pd[0]] - v[pd[1]])
		  + 0.1875*(-v[pd[4]] + v[pd[5]]) - 0.125*v[pd[6]]
		  + 0.375*v[pd[10]] + 0.75*v[pd[19]]);
	break;
      case 2:
	cdi = el->child[1]->dof[node0+1][n0];      
	TEST_EXIT(cdi == cd[17])("cdi != cd[17]\n");
	v[cdi] = (0.0625*(v[pd[0]] - v[pd[1]])
		  + 0.1875*(-v[pd[4]] + v[pd[5]]) - 0.125*v[pd[8]]
		  + 0.375*v[pd[12]] + 0.75*v[pd[18]]);
	break;
      }
    }
  }

  return;
}

static void real_coarse_inter3(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_inter3");
  EL        *el;
  REAL      *v = nil;
  int       i, lr_set;
  int       node_e, node_f, n0_e, n0_f;
  DOF       **pds, **cds;
  DOF       pd_o[N_BAS3], cd, pd;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  node_e = drv->fe_space->mesh->node[EDGE];        
  n0_e = admin->n0_dof[EDGE];
  node_f = drv->fe_space->mesh->node[FACE];        
  n0_f = admin->n0_dof[FACE];

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/
  pds = el->dof;
  cds = el->child[0]->dof;
  get_dof_indices(el, admin, pd_o);

  pd = (pds[0][0] < pds[1][0]) ? pds[node_e][n0_e] : pds[node_e][n0_e+1];
  cd = cds[0][0] < cds[3][0] ? cds[node_e+2][n0_e+1] : cds[node_e+2][n0_e];
  v[pd] = v[cd];

  pd = el->dof[node_f+2][n0_f];
  cd = cds[2][0] < cds[3][0] ? cds[node_e+5][n0_e+1] : cds[node_e+5][n0_e];
  v[pd] = v[cd];

  pd = el->dof[node_f+3][n0_f]; 
  cd = cds[1][0] < cds[3][0] ? cds[node_e+4][n0_e+1] : cds[node_e+4][n0_e];
  v[pd] = v[cd];

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cds = el->child[1]->dof;

  pd = (pds[0][0] < pds[1][0]) ? pds[node_e][n0_e+1] : pds[node_e][n0_e];
  cd = cds[0][0] < cds[3][0] ? cds[node_e+2][n0_e+1] : cds[node_e+2][n0_e];
  v[pd] = v[cd];

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/

  for (i = 1; i < n; i++)
  {
    el = list[i].el;

    pds = el->dof;
    cds = el->child[0]->dof;

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

    switch(lr_set)
    {
    case 1:
      pd = el->dof[node_f+3][n0_f];
      cd = cds[1][0] < cds[3][0] ? 
	cds[node_e+4][n0_e+1] : cds[node_e+4][n0_e];
      v[pd] = v[cd];
      break;
    case 2:
      pd = el->dof[node_f+2][n0_f];
      cd = cds[2][0] < cds[3][0] ?
	cds[node_e+5][n0_e+1] : cds[node_e+5][n0_e];
      v[pd] = v[cd];
      break;
    }
  }
  return;
}

static void real_coarse_restr3(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_restr3");
  EL        *el;
  REAL      *v = nil;
  const DOF *cd;
  DOF       pd[N_BAS3], cdi;
  int       node0, n0, i, typ, lr_set;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  v[pd[0]] += (0.0625*(-v[cd[3]] + v[cd[12]] + v[cd[14]]
		       + v[cd[16]] - v[cd[17]] - v[cd[18]]) 
	       + 0.3125*v[cd[8]]);
  v[pd[1]] += (0.0625*(-v[cd[3]] + v[cd[8]] + v[cd[12]] + v[cd[14]]
		       + v[cd[16]] + v[cd[17]] + v[cd[18]]));
  v[pd[4]] = (v[cd[9]] + 0.5625*v[cd[3]] + 0.9375*v[cd[8]]
	      + 0.0625*(-v[cd[12]] - v[cd[14]] - v[cd[16]])
	      + 0.1875*(v[cd[17]] + v[cd[18]]));
  v[pd[5]] = (0.5625*v[cd[3]] - 0.3125*v[cd[8]]
	      + 0.0625*(-v[cd[12]] - v[cd[14]] - v[cd[16]])
	      + 0.1875*(-v[cd[17]] - v[cd[18]]));
  v[pd[6]] += (-0.25*v[cd[12]] - 0.125*v[cd[16]] + 0.375*v[cd[18]]);
  v[pd[7]] += 0.5*v[cd[12]];
  v[pd[8]] += (-0.25*v[cd[14]] - 0.125*v[cd[16]] + 0.375*v[cd[17]]);
  v[pd[9]] += 0.5*v[cd[14]];
  v[pd[10]] += (-0.25*v[cd[12]] + 0.125*(-v[cd[16]] - v[cd[18]]));
  v[pd[11]] += 0.5*v[cd[12]];
  v[pd[12]] += (-0.25*v[cd[14]] + 0.125*(-v[cd[16]] - v[cd[17]]));
  v[pd[13]] += 0.5*v[cd[14]];
  v[pd[16]] += 0.5*v[cd[16]];
  v[pd[17]] += 0.5*v[cd[16]];
  v[pd[18]] = (v[cd[15]] + 0.5*v[cd[14]] + 0.25*v[cd[16]]
		 + 0.75*v[cd[17]]);
  v[pd[19]] = (v[cd[13]] + 0.5*v[cd[12]] + 0.25*v[cd[16]]
		 + 0.75*v[cd[18]]);

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);

  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    v[pd[0]] += 0.0625*(v[cd[8]] + v[cd[17]] + v[cd[18]]);
    v[pd[1]] += 0.3125*v[cd[8]] + 0.0625*(-v[cd[17]] - v[cd[18]]);
    v[pd[4]] += -0.3125*v[cd[8]] + 0.1875*(-v[cd[17]] - v[cd[18]]);
    v[pd[5]] += (v[cd[9]] + 0.9375*v[cd[8]] + 0.1875*(v[cd[17]] + v[cd[18]]));
    v[pd[6]] += -0.125*v[cd[17]];
    v[pd[8]] += -0.125*v[cd[18]];
    v[pd[10]] += 0.375*v[cd[17]];
    v[pd[12]] += 0.375*v[cd[18]];
    v[pd[18]] += 0.75*v[cd[18]];
    v[pd[19]] += 0.750000*v[cd[17]];
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 1|2                                                   */
/*--------------------------------------------------------------------------*/

    v[pd[0]] += 0.0625*(v[cd[8]] + v[cd[17]] + v[cd[18]]);
    v[pd[1]] += 0.3125*v[cd[8]] + 0.0625*(-v[cd[17]] - v[cd[18]]);
    v[pd[4]] += -0.3125*v[cd[8]] + 0.1875*(-v[cd[17]] - v[cd[18]]);
    v[pd[5]] += (v[cd[9]] + 0.9375*v[cd[8]] + 0.1875*(v[cd[17]] + v[cd[18]]));
    v[pd[6]] += -0.125*v[cd[18]];
    v[pd[8]] += -0.125*v[cd[17]];
    v[pd[10]] += 0.375*v[cd[18]];
    v[pd[12]] += 0.375*v[cd[17]];
    v[pd[18]] += 0.75*v[cd[17]];
    v[pd[19]] += 0.75*v[cd[18]];
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/

  node0 = drv->fe_space->mesh->node[FACE];
  n0 = admin->n0_dof[FACE];

  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);
    cdi = cd[16];

    switch(lr_set)
    {
    case 1:
      v[pd[0]] += 0.0625*(v[cd[12]] + v[cdi] - v[cd[18]]);
      v[pd[1]] += 0.0625*(v[cd[12]] + v[cdi] + v[cd[18]]);
      v[pd[4]] += (0.0625*(-v[cd[12]] - v[cdi]) + 0.1875*v[cd[18]]);
      v[pd[5]] += (0.0625*(-v[cd[12]] - v[cdi]) - 0.1875*v[cd[18]]);
      v[pd[6]] += (-0.25*v[cd[12]] - 0.125*v[cdi] + 0.375*v[cd[18]]);
      v[pd[7]] += 0.5*v[cd[12]];
      v[pd[8]] += -0.125*v[cdi];
      v[pd[10]] += -0.25*v[cd[12]] + 0.125*(-v[cdi] - v[cd[18]]);
      v[pd[11]] += 0.5*v[cd[12]];
      v[pd[12]] += -0.125*v[cdi];
      v[pd[16]] += 0.5*v[cdi];
      v[pd[17]] += 0.5*v[cdi];
      v[pd[18]] += 0.25*v[cdi];
      v[pd[19]] = (v[cd[13]] + 0.5*v[cd[12]] + 0.25*v[cdi] + 0.75*v[cd[18]]);
      break;
    case 2:
      v[pd[0]] += 0.0625*(v[cd[14]] + v[cdi] - v[cd[17]]);
      v[pd[1]] += 0.0625*(v[cd[14]] + v[cdi] + v[cd[17]]);
      v[pd[4]] += 0.0625*(-v[cd[14]] - v[cdi]) + 0.1875*v[cd[17]];
      v[pd[5]] += 0.0625*(-v[cd[14]] - v[cdi]) - 0.1875*v[cd[17]];
      v[pd[6]] += -0.125*v[cdi];
      v[pd[8]] += -0.25*v[cd[14]] - 0.125*v[cdi] + 0.375*v[cd[17]];
      v[pd[9]] += 0.5*v[cd[14]];
      v[pd[10]] += -0.125*v[cdi];
      v[pd[12]] += -0.25*v[cd[14]] + 0.125*(-v[cdi] - v[cd[17]]);
      v[pd[13]] += 0.5*v[cd[14]];
      v[pd[16]] += 0.5*v[cdi];
      v[pd[17]] += 0.5*v[cdi];
      v[pd[18]] = (v[cd[15]] + 0.5*v[cd[14]] + 0.25*v[cdi] + 0.75*v[cd[17]]);
      v[pd[19]] += 0.25*v[cdi];
      break;
    case 3:
      v[pd[0]] += 0.0625*v[cdi];
      v[pd[1]] += 0.0625*v[cdi];
      v[pd[4]] += -0.0625*v[cdi];
      v[pd[5]] += -0.0625*v[cdi];
      v[pd[6]] += -0.125*v[cdi];
      v[pd[8]] += -0.125*v[cdi];
      v[pd[10]] += -0.125*v[cdi];
      v[pd[12]] += -0.125*v[cdi];
      v[pd[16]] += 0.5*v[cdi];
      v[pd[17]] += 0.5*v[cdi];
      v[pd[18]] += 0.25*v[cdi];
      v[pd[19]] += 0.25*v[cdi];
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	cdi = el->child[1]->dof[node0+1][n0];      
	TEST_EXIT(cdi == cd[17])("cdi != cd[17]\n");
	v[pd[0]] += 0.0625*v[cdi];
	v[pd[1]] += -0.0625*v[cdi];
	v[pd[4]] += -0.1875*v[cdi];
	v[pd[5]] += 0.1875*v[cdi];
	v[pd[6]] += -0.125*v[cdi];
	v[pd[10]] += 0.375*v[cdi];
	v[pd[19]] += 0.75*v[cdi];
	break;
      case 2:
	cdi = el->child[1]->dof[node0+2][n0];      
	TEST_EXIT(cdi == cd[18])("cdi != cd[18]\n");
	v[pd[0]] += 0.0625*v[cdi];
	v[pd[1]] += -0.0625*v[cdi];
	v[pd[4]] += -0.1875*v[cdi];
	v[pd[5]] += 0.1875*v[cdi];
	v[pd[8]] += -0.125*v[cdi];
	v[pd[12]] += 0.375*v[cdi];
	v[pd[18]] += 0.75*v[cdi];
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	cdi = el->child[1]->dof[node0+2][n0];      
	TEST_EXIT(cdi == cd[18])("cdi != cd[18]\n");
	v[pd[0]] += 0.0625*v[cdi];
	v[pd[1]] += -0.0625*v[cdi];
	v[pd[4]] += -0.1875*v[cdi];
	v[pd[5]] += 0.1875*v[cdi];
	v[pd[6]] += -0.125*v[cdi];
	v[pd[10]] += 0.375*v[cdi];
	v[pd[19]] += 0.75*v[cdi];
	break;
      case 2:
	cdi = el->child[1]->dof[node0+1][n0];      
	TEST_EXIT(cdi == cd[17])("cdi != cd[17]\n");
	v[pd[0]] += 0.0625*v[cdi];
	v[pd[1]] += -0.0625*v[cdi];
	v[pd[4]] += -0.1875*v[cdi];
	v[pd[5]] += 0.1875*v[cdi];
	v[pd[8]] += -0.125*v[cdi];
	v[pd[12]] += 0.375*v[cdi];
	v[pd[18]] += 0.75*v[cdi];
	break;
      }
    }
  }
  return;
}


static void real_d_refine_inter3(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_refine_inter3");
  EL        *el;
  REAL_D    *v = nil;
  const DOF *cd;
  DOF       pd[N_BAS3], cdi;
  int       node0, n0, i, typ, lr_set, k;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  for (k = 0; k < DIM_OF_WORLD; k++)
  {
    v[cd[3]][k] = (0.0625*(-v[pd[0]][k] - v[pd[1]][k])
		   + 0.5625*(v[pd[4]][k] + v[pd[5]][k]));
    v[cd[8]][k] = (0.3125*(v[pd[0]][k] - v[pd[5]][k]) + 0.0625*v[pd[1]][k]
		   + 0.937500*v[pd[4]][k]);
    v[cd[9]][k] = v[pd[4]][k];
    v[cd[12]][k] = (0.0625*(v[pd[0]][k] + v[pd[1]][k]
			    - v[pd[4]][k] - v[pd[5]][k])
		    + 0.25*(-v[pd[6]][k] - v[pd[10]][k])
		    + 0.5*(v[pd[7]][k] + v[pd[11]][k] + v[pd[19]][k]));
    v[cd[13]][k] = v[pd[19]][k];
    v[cd[14]][k] = (0.0625*(v[pd[0]][k] + v[pd[1]][k] - v[pd[4]][k]
			    - v[pd[5]][k]) 
		    + 0.25*(-v[pd[8]][k] - v[pd[12]][k]) 
		    + 0.5*(v[pd[9]][k] + v[pd[13]][k] + v[pd[18]][k]));
    v[cd[15]][k] = (v[pd[18]][k]);
    v[cd[16]][k] = (0.0625*(v[pd[0]][k] + v[pd[1]][k]
			    - v[pd[4]][k] - v[pd[5]][k])
		    + 0.125*(-v[pd[6]][k] - v[pd[8]][k]
			     - v[pd[10]][k] - v[pd[12]][k])
		    + 0.5*(v[pd[16]][k] + v[pd[17]][k]) 
		    + 0.25*(v[pd[18]][k] + v[pd[19]][k]));
    v[cd[17]][k] = (0.0625*(-v[pd[0]][k] + v[pd[1]][k]) 
		    + 0.1875*(v[pd[4]][k] - v[pd[5]][k]) + 0.375*v[pd[8]][k]
		    - 0.125*v[pd[12]][k] + 0.75*v[pd[18]][k]);
    v[cd[18]][k] = (0.0625*(-v[pd[0]][k] + v[pd[1]][k]) 
		    + 0.1875*(v[pd[4]][k] - v[pd[5]][k]) + 0.375*v[pd[6]][k]
		    - 0.125*v[pd[10]][k] + 0.75*v[pd[19]][k]);
  }
/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);

  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[cd[8]][k] = (0.0625*v[pd[0]][k] + 0.3125*(v[pd[1]][k] - v[pd[4]][k])
		     + 0.9375*v[pd[5]][k]);
      v[cd[9]][k] = v[pd[5]][k];
      v[cd[17]][k] = (0.0625*(v[pd[0]][k] - v[pd[1]][k]) 
		      + 0.1875*(-v[pd[4]][k] + v[pd[5]][k])
		      - 0.125*v[pd[6]][k] + 0.375*v[pd[10]][k]
		      + 0.75*v[pd[19]][k]);
      v[cd[18]][k] = (0.0625*(v[pd[0]][k] - v[pd[1]][k]) 
		      + 0.1875*(-v[pd[4]][k] + v[pd[5]][k])
		      - 0.125*v[pd[8]][k] 
		      + 0.375*v[pd[12]][k] + 0.75*v[pd[18]][k]);
    }
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[cd[8]][k] = (0.0625*v[pd[0]][k] + 0.3125*(v[pd[1]][k] - v[pd[4]][k])
		     + 0.9375*v[pd[5]][k]);
      v[cd[9]][k] = v[pd[5]][k];
      v[cd[17]][k] = (0.0625*(v[pd[0]][k] - v[pd[1]][k]) 
		      + 0.1875*(-v[pd[4]][k] + v[pd[5]][k])
		      - 0.125*v[pd[8]][k] + 0.375*v[pd[12]][k]
		      + 0.75*v[pd[18]][k]);
      v[cd[18]][k] = (0.0625*(v[pd[0]][k] - v[pd[1]][k]) + 
		      0.1875*(-v[pd[4]][k] + v[pd[5]][k]) - 0.125*v[pd[6]][k]
		      + 0.375*v[pd[10]][k] + 0.75*v[pd[19]][k]);
    }
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/

  node0 = drdv->fe_space->mesh->node[FACE];
  n0 = admin->n0_dof[FACE];
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[cd[12]][k] = (0.0625*(v[pd[0]][k]+v[pd[1]][k]
				-v[pd[4]][k]-v[pd[5]][k])
			+ 0.25*(-v[pd[6]][k] - v[pd[10]][k]) 
			+ 0.5*(v[pd[7]][k] + v[pd[11]][k] + v[pd[19]][k]));
	v[cd[13]][k] = v[pd[19]][k];
	v[cd[16]][k] = (0.0625*(v[pd[0]][k]+v[pd[1]][k]
				-v[pd[4]][k]-v[pd[5]][k])
			+ 0.125*(-v[pd[6]][k]-v[pd[8]][k]
				 -v[pd[10]][k]-v[pd[12]][k])
			+ 0.5*(v[pd[16]][k] + v[pd[17]][k]) 
			+ 0.25*(v[pd[18]][k] + v[pd[19]][k]));
	v[cd[18]][k] = (0.0625*(-v[pd[0]][k] + v[pd[1]][k])
			+ 0.1875*(v[pd[4]][k] - v[pd[5]][k])
			+ 0.375*v[pd[6]][k] - 0.125*v[pd[10]][k]
			+ 0.75*v[pd[19]][k]);
      }
      break;
    case 2:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[cd[14]][k] = (0.0625*(v[pd[0]][k]+v[pd[1]][k]
				-v[pd[4]][k]-v[pd[5]][k])
			+ 0.25*(-v[pd[8]][k] - v[pd[12]][k]) 
			+ 0.5*(v[pd[9]][k] + v[pd[13]][k] + v[pd[18]][k]));
	v[cd[15]][k] = v[pd[18]][k];
	v[cd[16]][k] = (0.0625*(v[pd[0]][k]+v[pd[1]][k]
				-v[pd[4]][k]-v[pd[5]][k])
			+ 0.125*(-v[pd[6]][k]-v[pd[8]][k]
				 -v[pd[10]][k]-v[pd[12]][k])
			+ 0.5*(v[pd[16]][k] + v[pd[17]][k])
			+ 0.25*(v[pd[18]][k] + v[pd[19]][k]));
	v[cd[17]][k] = (0.0625*(-v[pd[0]][k] + v[pd[1]][k])
			+ 0.1875*(v[pd[4]][k] - v[pd[5]][k])
			+ 0.375*v[pd[8]][k] - 0.125*v[pd[12]][k]
			+ 0.75*v[pd[18]][k]);
      }
      break;
    case 3:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {    
	v[cd[16]][k] = (0.0625*(v[pd[0]][k]+v[pd[1]][k]
				-v[pd[4]][k]-v[pd[5]][k])
			+ 0.125*(-v[pd[6]][k]-v[pd[8]][k]
				 -v[pd[10]][k]-v[pd[12]][k])
			+ 0.5*(v[pd[16]][k] + v[pd[17]][k])
			+ 0.25*(v[pd[18]][k] + v[pd[19]][k]));
      }
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	cdi = el->child[1]->dof[node0+1][n0];      
	TEST_EXIT(cdi == cd[17])("cdi != cd[17]\n");
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cdi][k] = (0.0625*(v[pd[0]][k] - v[pd[1]][k])
		       + 0.1875*(-v[pd[4]][k] + v[pd[5]][k])
		       - 0.125*v[pd[6]][k] + 0.375*v[pd[10]][k]
		       + 0.75*v[pd[19]][k]);
	}
	break;
      case 2:
	cdi = el->child[1]->dof[node0+2][n0];      
	TEST_EXIT(cdi == cd[18])("cdi != cd[18]\n");
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cdi][k] = (0.0625*(v[pd[0]][k] - v[pd[1]][k])
		       + 0.1875*(-v[pd[4]][k] + v[pd[5]][k])
		       - 0.125*v[pd[8]][k]
		       + 0.375*v[pd[12]][k] + 0.75*v[pd[18]][k]);
	}
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	cdi = el->child[1]->dof[node0+2][n0];      
	TEST_EXIT(cdi == cd[18])("cdi != cd[18]\n");
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cdi][k] = (0.0625*(v[pd[0]][k] - v[pd[1]][k])
		       + 0.1875*(-v[pd[4]][k] + v[pd[5]][k])
		       - 0.125*v[pd[6]][k] + 0.375*v[pd[10]][k]
		       + 0.75*v[pd[19]][k]);
	}
	break;
      case 2:
	cdi = el->child[1]->dof[node0+1][n0];      
	TEST_EXIT(cdi == cd[17])("cdi != cd[17]\n");
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cdi][k] = (0.0625*(v[pd[0]][k] - v[pd[1]][k])
		       + 0.1875*(-v[pd[4]][k] + v[pd[5]][k])
		       - 0.125*v[pd[8]][k] + 0.375*v[pd[12]][k]
		       + 0.75*v[pd[18]][k]);
	}
	break;
      }
    }
  }

  return;
}

static void real_d_coarse_inter3(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_inter3");
  EL        *el;
  REAL_D    *v = nil;
  int       i, lr_set, k;
  int       node_e, node_f, n0_e, n0_f;
  DOF       **pds, **cds;
  DOF       pd_o[N_BAS3], cd, pd;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  node_e = drdv->fe_space->mesh->node[EDGE];        
  n0_e = admin->n0_dof[EDGE];
  node_f = drdv->fe_space->mesh->node[FACE];        
  n0_f = admin->n0_dof[FACE];

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/
  pds = el->dof;
  cds = el->child[0]->dof;
  get_dof_indices(el, admin, pd_o);

  pd = (pds[0][0] < pds[1][0]) ? pds[node_e][n0_e] : pds[node_e][n0_e+1];
  cd = cds[0][0] < cds[3][0] ? cds[node_e+2][n0_e+1] : cds[node_e+2][n0_e];
  for (k = 0; k < DIM_OF_WORLD; k++)
    v[pd][k] = v[cd][k];

  pd = el->dof[node_f+2][n0_f];
  cd = cds[2][0] < cds[3][0] ? cds[node_e+5][n0_e+1] : cds[node_e+5][n0_e];
  for (k = 0; k < DIM_OF_WORLD; k++)
    v[pd][k] = v[cd][k];

  pd = el->dof[node_f+3][n0_f]; 
  cd = cds[1][0] < cds[3][0] ? cds[node_e+4][n0_e+1] : cds[node_e+4][n0_e];
  for (k = 0; k < DIM_OF_WORLD; k++)
    v[pd][k] = v[cd][k];

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cds = el->child[1]->dof;

  pd = (pds[0][0] < pds[1][0]) ? pds[node_e][n0_e+1] : pds[node_e][n0_e];
  cd = cds[0][0] < cds[3][0] ? cds[node_e+2][n0_e+1] : cds[node_e+2][n0_e];
  for (k = 0; k < DIM_OF_WORLD; k++)
    v[pd][k] = v[cd][k];


/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/

  for (i = 1; i < n; i++)
  {
    el = list[i].el;

    pds = el->dof;
    cds = el->child[0]->dof;

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

    switch(lr_set)
    {
    case 1:
      pd = el->dof[node_f+3][n0_f];
      cd = cds[1][0] < cds[3][0] ? 
	cds[node_e+4][n0_e+1] : cds[node_e+4][n0_e];
      for (k = 0; k < DIM_OF_WORLD; k++)
	v[pd][k] = v[cd][k];
      break;
    case 2:
      pd = el->dof[node_f+2][n0_f];
      cd = cds[2][0] < cds[3][0] ?
	cds[node_e+5][n0_e+1] : cds[node_e+5][n0_e];
      for (k = 0; k < DIM_OF_WORLD; k++)
	v[pd][k] = v[cd][k];
      break;
    }
  }
  return;
}

static void real_d_coarse_restr3(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_restr3");
  EL        *el;
  REAL_D    *v = nil;
  const DOF *cd;
  DOF       pd[N_BAS3], cdi;
  int       node0, n0, i, k, typ, lr_set;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  for (k = 0; k < DIM_OF_WORLD; k++)
  {
    v[pd[0]][k] += (0.0625*(-v[cd[3]][k] + v[cd[12]][k] + v[cd[14]][k]
			    + v[cd[16]][k] - v[cd[17]][k] - v[cd[18]][k]) 
		    + 0.3125*v[cd[8]][k]);
    v[pd[1]][k] += (0.0625*(-v[cd[3]][k] + v[cd[8]][k] + v[cd[12]][k]
			    + v[cd[14]][k] + v[cd[16]][k] + v[cd[17]][k]
			    + v[cd[18]][k]));
    v[pd[4]][k] = (v[cd[9]][k] + 0.5625*v[cd[3]][k] + 0.9375*v[cd[8]][k]
		   + 0.0625*(-v[cd[12]][k] - v[cd[14]][k] - v[cd[16]][k])
		   + 0.1875*(v[cd[17]][k] + v[cd[18]][k]));
    v[pd[5]][k] = (0.5625*v[cd[3]][k] - 0.3125*v[cd[8]][k]
		   + 0.0625*(-v[cd[12]][k] - v[cd[14]][k] - v[cd[16]][k])
		   + 0.1875*(-v[cd[17]][k] - v[cd[18]][k]));
    v[pd[6]][k] += (-0.25*v[cd[12]][k] - 0.125*v[cd[16]][k]
		    + 0.375*v[cd[18]][k]);
    v[pd[7]][k] += 0.5*v[cd[12]][k];
    v[pd[8]][k] += (-0.25*v[cd[14]][k] - 0.125*v[cd[16]][k]
		    + 0.375*v[cd[17]][k]);
    v[pd[9]][k] += 0.5*v[cd[14]][k];
    v[pd[10]][k] += (-0.25*v[cd[12]][k]
		     + 0.125*(-v[cd[16]][k] - v[cd[18]][k]));
    v[pd[11]][k] += 0.5*v[cd[12]][k];
    v[pd[12]][k] += (-0.25*v[cd[14]][k]
		     + 0.125*(-v[cd[16]][k] - v[cd[17]][k]));
    v[pd[13]][k] += 0.5*v[cd[14]][k];
    v[pd[16]][k] += 0.5*v[cd[16]][k];
    v[pd[17]][k] += 0.5*v[cd[16]][k];
    v[pd[18]][k] = (v[cd[15]][k] + 0.5*v[cd[14]][k] + 0.25*v[cd[16]][k]
		    + 0.75*v[cd[17]][k]);
    v[pd[19]][k] = (v[cd[13]][k] + 0.5*v[cd[12]][k] + 0.25*v[cd[16]][k]
		    + 0.75*v[cd[18]][k]);
  }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);

  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    for (k = 0; k < DIM_OF_WORLD; k++)
    {   
      v[pd[0]][k] += 0.0625*(v[cd[8]][k] + v[cd[17]][k] + v[cd[18]][k]);
      v[pd[1]][k] += (0.3125*v[cd[8]][k] 
		      + 0.0625*(-v[cd[17]][k] - v[cd[18]][k]));
      v[pd[4]][k] += (-0.3125*v[cd[8]][k]
		      + 0.1875*(-v[cd[17]][k] - v[cd[18]][k]));
      v[pd[5]][k] += (v[cd[9]][k] + 0.9375*v[cd[8]][k]
		      + 0.1875*(v[cd[17]][k] + v[cd[18]][k]));
      v[pd[6]][k] += -0.125*v[cd[17]][k];
      v[pd[8]][k] += -0.125*v[cd[18]][k];
      v[pd[10]][k] += 0.375*v[cd[17]][k];
      v[pd[12]][k] += 0.375*v[cd[18]][k];
      v[pd[18]][k] += 0.75*v[cd[18]][k];
      v[pd[19]][k] += 0.750000*v[cd[17]][k];
    }
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 1|2                                                   */
/*--------------------------------------------------------------------------*/

    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[pd[0]][k] += 0.0625*(v[cd[8]][k] + v[cd[17]][k] + v[cd[18]][k]);
      v[pd[1]][k] += (0.3125*v[cd[8]][k]
		      + 0.0625*(-v[cd[17]][k] - v[cd[18]][k]));
      v[pd[4]][k] += (-0.3125*v[cd[8]][k]
		      + 0.1875*(-v[cd[17]][k] - v[cd[18]][k]));
      v[pd[5]][k] += (v[cd[9]][k] + 0.9375*v[cd[8]][k]
		      + 0.1875*(v[cd[17]][k] + v[cd[18]][k]));
      v[pd[6]][k] += -0.125*v[cd[18]][k];
      v[pd[8]][k] += -0.125*v[cd[17]][k];
      v[pd[10]][k] += 0.375*v[cd[18]][k];
      v[pd[12]][k] += 0.375*v[cd[17]][k];
      v[pd[18]][k] += 0.75*v[cd[17]][k];
      v[pd[19]][k] += 0.75*v[cd[18]][k];
    }
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/

  node0 = drdv->fe_space->mesh->node[FACE];
  n0 = admin->n0_dof[FACE];

  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);
    cdi = cd[16];

    switch(lr_set)
    {
    case 1:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[0]][k] += 0.0625*(v[cd[12]][k] + v[cdi][k] - v[cd[18]][k]);
	v[pd[1]][k] += 0.0625*(v[cd[12]][k] + v[cdi][k] + v[cd[18]][k]);
	v[pd[4]][k] += (0.0625*(-v[cd[12]][k] - v[cdi][k])
			+ 0.1875*v[cd[18]][k]);
	v[pd[5]][k] += (0.0625*(-v[cd[12]][k] - v[cdi][k])
			- 0.1875*v[cd[18]][k]);
	v[pd[6]][k] += (-0.25*v[cd[12]][k] - 0.125*v[cdi][k]
			+ 0.375*v[cd[18]][k]);
	v[pd[7]][k] += 0.5*v[cd[12]][k];
	v[pd[8]][k] += -0.125*v[cdi][k];
	v[pd[10]][k] += -0.25*v[cd[12]][k] + 0.125*(-v[cdi][k]-v[cd[18]][k]);
	v[pd[11]][k] += 0.5*v[cd[12]][k];
	v[pd[12]][k] += -0.125*v[cdi][k];
	v[pd[16]][k] += 0.5*v[cdi][k];
	v[pd[17]][k] += 0.5*v[cdi][k];
	v[pd[18]][k] += 0.25*v[cdi][k];
	v[pd[19]][k] = (v[cd[13]][k] + 0.5*v[cd[12]][k] + 0.25*v[cdi][k]
			+ 0.75*v[cd[18]][k]);
      }
      break;
    case 2:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[0]][k] += 0.0625*(v[cd[14]][k] + v[cdi][k] - v[cd[17]][k]);
	v[pd[1]][k] += 0.0625*(v[cd[14]][k] + v[cdi][k] + v[cd[17]][k]);
	v[pd[4]][k] += (0.0625*(-v[cd[14]][k] - v[cdi][k])
			+ 0.1875*v[cd[17]][k]);
	v[pd[5]][k] += (0.0625*(-v[cd[14]][k] - v[cdi][k])
			- 0.1875*v[cd[17]][k]);
	v[pd[6]][k] += -0.125*v[cdi][k];
	v[pd[8]][k] += (-0.25*v[cd[14]][k] - 0.125*v[cdi][k]
			+ 0.375*v[cd[17]][k]);
	v[pd[9]][k] += 0.5*v[cd[14]][k];
	v[pd[10]][k] += -0.125*v[cdi][k];
	v[pd[12]][k] += -0.25*v[cd[14]][k] + 0.125*(-v[cdi][k]-v[cd[17]][k]);
	v[pd[13]][k] += 0.5*v[cd[14]][k];
	v[pd[16]][k] += 0.5*v[cdi][k];
	v[pd[17]][k] += 0.5*v[cdi][k];
	v[pd[18]][k] = (v[cd[15]][k] + 0.5*v[cd[14]][k]
			+ 0.25*v[cdi][k] + 0.75*v[cd[17]][k]);
	v[pd[19]][k] += 0.25*v[cdi][k];
      }
      break;
    case 3:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[0]][k] += 0.0625*v[cdi][k];
	v[pd[1]][k] += 0.0625*v[cdi][k];
	v[pd[4]][k] += -0.0625*v[cdi][k];
	v[pd[5]][k] += -0.0625*v[cdi][k];
	v[pd[6]][k] += -0.125*v[cdi][k];
	v[pd[8]][k] += -0.125*v[cdi][k];
	v[pd[10]][k] += -0.125*v[cdi][k];
	v[pd[12]][k] += -0.125*v[cdi][k];
	v[pd[16]][k] += 0.5*v[cdi][k];
	v[pd[17]][k] += 0.5*v[cdi][k];
	v[pd[18]][k] += 0.25*v[cdi][k];
	v[pd[19]][k] += 0.25*v[cdi][k];
      }
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	cdi = el->child[1]->dof[node0+1][n0];      
	TEST_EXIT(cdi == cd[17])("cdi != cd[17]\n");
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += 0.0625*v[cdi][k];
	  v[pd[1]][k] += -0.0625*v[cdi][k];
	  v[pd[4]][k] += -0.1875*v[cdi][k];
	  v[pd[5]][k] += 0.1875*v[cdi][k];
	  v[pd[6]][k] += -0.125*v[cdi][k];
	  v[pd[10]][k] += 0.375*v[cdi][k];
	  v[pd[19]][k] += 0.75*v[cdi][k];
	}
	break;
      case 2:
	cdi = el->child[1]->dof[node0+2][n0];      
	TEST_EXIT(cdi == cd[18])("cdi != cd[18]\n");
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += 0.0625*v[cdi][k];
	  v[pd[1]][k] += -0.0625*v[cdi][k];
	  v[pd[4]][k] += -0.1875*v[cdi][k];
	  v[pd[5]][k] += 0.1875*v[cdi][k];
	  v[pd[8]][k] += -0.125*v[cdi][k];
	  v[pd[12]][k] += 0.375*v[cdi][k];
	  v[pd[18]][k] += 0.75*v[cdi][k];
	}
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	cdi = el->child[1]->dof[node0+2][n0];      
	TEST_EXIT(cdi == cd[18])("cdi != cd[18]\n");
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += 0.0625*v[cdi][k];
	  v[pd[1]][k] += -0.0625*v[cdi][k];
	  v[pd[4]][k] += -0.1875*v[cdi][k];
	  v[pd[5]][k] += 0.1875*v[cdi][k];
	  v[pd[6]][k] += -0.125*v[cdi][k];
	  v[pd[10]][k] += 0.375*v[cdi][k];
	  v[pd[19]][k] += 0.75*v[cdi][k];
	}
	break;
      case 2:
	cdi = el->child[1]->dof[node0+1][n0];      
	TEST_EXIT(cdi == cd[17])("cdi != cd[17]\n");
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += 0.0625*v[cdi][k];
	  v[pd[1]][k] += -0.0625*v[cdi][k];
	  v[pd[4]][k] += -0.1875*v[cdi][k];
	  v[pd[5]][k] += 0.1875*v[cdi][k];
	  v[pd[8]][k] += -0.125*v[cdi][k];
	  v[pd[12]][k] += 0.375*v[cdi][k];
	  v[pd[18]][k] += 0.75*v[cdi][k];
	}
	break;
      }
    }
  }
  return;
}

static BAS_FCT      *phi3[N_BAS3]     = {phi3v0,  phi3v1, phi3v2,  phi3v3,
					 phi3e0,  phi3e1, phi3e2,  phi3e3,
					 phi3e4,  phi3e5, phi3e6,  phi3e7,
					 phi3e8,  phi3e9, phi3e10, phi3e11,
					 phi3c0,  phi3c1, phi3c2,  phi3c3};

static GRD_BAS_FCT  *grd_phi3[N_BAS3] = {grd_phi3v0,  grd_phi3v1,
					 grd_phi3v2,  grd_phi3v3,
					 grd_phi3e0,  grd_phi3e1,
					 grd_phi3e2,  grd_phi3e3,
					 grd_phi3e4,  grd_phi3e5,
					 grd_phi3e6,  grd_phi3e7,
					 grd_phi3e8,  grd_phi3e9,
					 grd_phi3e10, grd_phi3e11,
					 grd_phi3c0,  grd_phi3c1,
					 grd_phi3c2,  grd_phi3c3};

static D2_BAS_FCT   *D2_phi3[N_BAS3]  = {D2_phi3v0,  D2_phi3v1,
					 D2_phi3v2,  D2_phi3v3,
					 D2_phi3e0,  D2_phi3e1,
					 D2_phi3e2,  D2_phi3e3,
					 D2_phi3e4,  D2_phi3e5,
					 D2_phi3e6,  D2_phi3e7,
					 D2_phi3e8,  D2_phi3e9,
					 D2_phi3e10, D2_phi3e11,
					 D2_phi3c0,  D2_phi3c1,
					 D2_phi3c2,  D2_phi3c3};

static const BAS_FCTS lagrange3 = {"lagrange3", N_BAS3, 3, {1, 2, 1, 0},
				   nil, phi3, grd_phi3, D2_phi3, 
				   get_dof_indices3, 
				   get_bound3,
				   interpol3,
				   interpol_d3,
				   get_int_vec3,
				   get_real_vec3,
				   get_real_d_vec3,
				   get_uchar_vec3,
				   get_schar_vec3,
				   real_refine_inter3,
				   real_coarse_inter3,
				   real_coarse_restr3,
				   real_d_refine_inter3,
				   real_d_coarse_inter3,
				   real_d_coarse_restr3,
				   (void *) bary3};
