/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/* file: graphics.c                                                         */
/*                                                                          */
/*                                                                          */
/* description: graphical output for mesh, discrete solution and estimate   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include <alberta.h>


#define USE_GLTOOLS (HAVE_LIBGLTOOLS  && true)

#if !USE_GLTOOLS
/*---8<---------------------------------------------------------------------*/
/*---   simple GL graphics ...                                           ---*/
/*---   nothing will be done in 3d                                       ---*/
/*--------------------------------------------------------------------->8---*/

void graphics(MESH *mesh, DOF_REAL_VEC *u_h, REAL (*get_est)(EL *el))
{
  FUNCNAME("graphics");
  static int first = true;
  static GRAPH_WINDOW win_est = nil, win_val = nil, win_mesh = nil;
  static REAL  min=0.0, max=-1.0;
  int    refine;

#if DIM < 3 
  if (first)
  {
    int size[3] = {0, 0, 0};
    char   geom[128] = "500x500+0+0";
    GET_PARAMETER(1, "graphic windows", "%d %d %d", size, size+1, size+2);
    GET_PARAMETER(1, "graphic range", "%e %e", &min, &max);

    if (size[0] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[0], size[0], 0, 0); 
      win_val = graph_open_window("ALBERTA values", geom, nil, mesh);
    }
    if (size[1] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[0], size[0], size[0], 0); 
      win_est = graph_open_window("ALBERTA estimate", geom, nil, mesh);
    }
    if (size[2] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[2], size[2], size[0]+size[1], 0); 
      win_mesh = graph_open_window("ALBERTA mesh", geom, nil, mesh);
    }
    first = false;
  }

  if (mesh && win_mesh)
  {
    graph_clear_window(win_mesh, rgb_white);
    graph_mesh(win_mesh, mesh, rgb_black, 0);
  }
  if (u_h && win_val)
  {
    graph_clear_window(win_val, rgb_white);
    refine = u_h->fe_space->bas_fcts->degree;
    if (refine<2) refine=0;
    graph_drv(win_val, u_h, min, max, refine);
  }

  if (get_est && win_est)
  {
    graph_clear_window(win_est, rgb_white);
    graph_el_est(win_est, mesh, get_est, 0.0, 0.0); 
  }
#endif
  WAIT;
  return;
}
#else

void graphics(MESH *mesh, DOF_REAL_VEC *u_h, REAL (*get_est)(EL *el))
{
  FUNCNAME("graphics");
  static int first = true;
  static GLTOOLS_WINDOW   win_est = nil, win_val = nil, win_mesh = nil;
  static REAL  min=0.0, max=-1.0;

  if (first)
  {
    int size[3] = {0, 0, 0};
    char   geom[128] = "500x500+0+0";
    GET_PARAMETER(1, "graphic windows", "%d %d %d", size, size+1, size+2);

    GET_PARAMETER(1, "graphic range", "%e %e", &min, &max);

    if (size[0] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[0], size[0], 0, 0); 
      win_val = open_gltools_window("ALBERTA values", geom, nil, mesh, true);
    }

    if (size[1] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[1], size[1], size[0], 0); 
      win_est = open_gltools_window("ALBERTA estimate", geom, nil, mesh, true);
    }

    if (size[2] > 0)
    {
      sprintf(geom, "%dx%d+%d+%d", size[2], size[2], size[0]+size[1], 0); 
      win_mesh = open_gltools_window("ALBERTA mesh", geom, nil, mesh, true);
    }
    first = false;
    if (!(u_h && win_val) && !(get_est && win_est) && !(mesh &&win_mesh))
	return;
  }


  if (mesh && win_mesh) {
    gltools_mesh(win_mesh, mesh, 0);
  }


  if (u_h && win_val) {
    gltools_drv(win_val, u_h, min, max);
  }

  if (win_est)
  {
    if (get_est)
      gltools_est(win_est, mesh, get_est, 0.0, -1.0);
/*     else if (mesh) */
/*       gltools_mesh(win_est, mesh, 0); */
  }


  if (!win_val && !win_est)  WAIT;

  return;
}
#endif
